import copy
from typing import List

from ..layers import DoNothingOptimizer
from ..utils import common_functions as c_f
from .base_container import BaseContainer


class Optimizers(BaseContainer):
    """
    A container for model optimizers.
    """

    def __init__(self, *args, multipliers=None, **kwargs):
        """
        Arguments:
            *args: [```BaseContainer```][pytorch_adapt.containers.BaseContainer] arguments.
            multipliers: A dictionary mapping from
                optimizer name to lr multiplier. Each
                optimizer will have ```lr = lr * multiplier```
                upon initialization. If ```None```,
                then multiplier is 1.
            **kwargs:  [```BaseContainer```][pytorch_adapt.containers.BaseContainer]
                keyword arguments.
        """
        self.multipliers = c_f.default(multipliers, {})
        super().__init__(*args, **kwargs)

    def _create_with(self, other):
        c_f.assert_keys_are_present_cls(self, "multipliers", self)
        for k, v in self.items():
            if c_f.is_optimizer(v):
                continue
            class_ref, kwargs = v
            model = other[k]
            if c_f.has_no_parameters(model):
                self[k] = DoNothingOptimizer()
            else:
                kwargs = copy.deepcopy(kwargs)
                kwargs["lr"] *= self.multipliers.get(k, 1)
                self[k] = class_ref(model.parameters(), **kwargs)

    def step(self):
        """
        Calls ```.step()``` on all optimizers.
        """
        for v in self.values():
            v.step()

    def zero_grad(self):
        """
        Calls ```.zero_grad()``` on all optimizers.
        """
        for v in self.values():
            v.zero_grad()

    def merge(self, other):
        super().merge(other)
        self.multipliers.update(other.multipliers)

    def zero_back_step(self, loss, keys: List[str] = None):
        """
        Zeros gradients, computes gradients, and updates model weights.
        Arguments:
            loss: The loss on which ```.backward()``` is called.
            keys: The subset of optimizers on which to call
                ```.zero_grad()``` and ```.step()```.
                If ```None```, then all optimizers are used.
        """
        keys = c_f.default(keys, self.keys())
        optimizers = [self[k] for k in keys]
        c_f.zero_back_step(loss, optimizers)



from typing import Tuple

import torch
import torch.nn.functional as F
from pytorch_metric_learning.utils import common_functions as pml_cf

from ..utils import common_functions as c_f


# reference https://github.com/tim-learn/ATDOC/blob/main/demo_uda.py
class NeighborhoodAggregation(torch.nn.Module):
    """
    Implementation of the pseudo labeling step in
    [Domain Adaptation with Auxiliary Target Domain-Oriented Classifier](https://arxiv.org/abs/2007.04171).
    """

    def __init__(
        self,
        dataset_size: int,
        feature_dim: int,
        num_classes: int,
        k: int = 5,
        T: float = 0.5,
    ):
        """
        Arguments:
            dataset_size: The number of samples in the target dataset.
            feature_dim: The feature dimensionality, i.e at each iteration
                the features should be size ```(N, D)``` where N is batch size and
                D is ```feature_dim```.
            num_classes: The number of class labels in the target dataset.
            k: The number of nearest neighbors used to determine each
                sample's pseudolabel
            T: The softmax temperature used when storing predictions in memory.
        """

        super().__init__()
        self.register_buffer(
            "feat_memory", F.normalize(torch.rand(dataset_size, feature_dim))
        )
        self.register_buffer(
            "pred_memory", torch.ones(dataset_size, num_classes) / num_classes
        )
        self.k = k
        self.T = T

    def forward(
        self,
        features: torch.Tensor,
        logits: torch.Tensor = None,
        update: bool = False,
        idx: torch.Tensor = None,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Arguments:
            features: The features to compute pseudolabels for.
            logits: The logits from which predictions will be computed and
                stored in memory. Required if ```update = True```
            update: If True, the current batch of predictions is
                added to the memory bank.
            idx: A tensor containing the dataset indices that
                produced each row of ```features```.
        """
        # move to device if necessary
        self.feat_memory = pml_cf.to_device(self.feat_memory, features)
        self.pred_memory = pml_cf.to_device(self.pred_memory, features)
        with torch.no_grad():
            features = F.normalize(features)
            pseudo_labels, mean_preds = self.get_pseudo_labels(features, idx)
            if update:
                self.update_memory(features, logits, idx)
        return pseudo_labels, mean_preds

    def get_pseudo_labels(self, normalized_features, idx):
        dis = torch.mm(normalized_features, self.feat_memory.t())
        # set self-comparisons to min similarity
        for di in range(dis.size(0)):
            dis[di, idx[di]] = torch.min(dis)
        _, indices = torch.topk(dis, k=self.k, dim=1)
        preds = torch.mean(self.pred_memory[indices], dim=1)
        pseudo_labels = torch.argmax(preds, dim=1)
        return pseudo_labels, preds

    def update_memory(self, normalized_features, logits, idx):
        preds = F.softmax(logits, dim=1)
        p = 1.0 / self.T
        preds = (preds**p) / torch.sum(preds**p, dim=0)
        self.feat_memory[idx] = normalized_features
        self.pred_memory[idx] = preds

    def extra_repr(self):
        """"""
        return c_f.extra_repr(self, ["k", "T"])



from typing import Callable, List

import torch

from ..utils import common_functions as c_f
from .base import BaseHook, BaseWrapperHook
from .features import (
    FeaturesAndLogitsHook,
    FeaturesChainHook,
    FeaturesHook,
    FrozenModelHook,
    LogitsHook,
)
from .optimizer import OptimizerHook, SummaryHook
from .utils import ApplyFnHook, ChainHook, OnlyNewOutputsHook


class SoftmaxHook(ApplyFnHook):
    """
    Applies ```torch.nn.Softmax(dim=1)``` to the
    specified inputs.
    """

    def __init__(self, **kwargs):
        super().__init__(fn=torch.nn.Softmax(dim=1), **kwargs)


class SoftmaxLocallyHook(BaseWrapperHook):
    """
    Applies ```torch.nn.Softmax(dim=1)``` to the
    specified inputs, which are overwritten, but
    only inside this hook.
    """

    def __init__(self, apply_to: List[str], *hooks: BaseHook, **kwargs):
        """
        Arguments:
            apply_to: list of names of tensors that softmax
                will be applied to.
            *hooks: the hooks that will receive the softmaxed
                tensors.
        """
        super().__init__(**kwargs)
        s_hook = SoftmaxHook(apply_to=apply_to)
        self.hook = OnlyNewOutputsHook(ChainHook(s_hook, *hooks, overwrite=True))


class CLossHook(BaseWrapperHook):
    """
    Computes a classification loss on the specified tensors.
    The default setting is to compute the cross entropy loss
    of the source domain logits.
    """

    def __init__(
        self,
        loss_fn: Callable[[torch.Tensor, torch.Tensor], torch.Tensor] = None,
        detach_features: bool = False,
        f_hook: BaseHook = None,
        domains=("src",),
        **kwargs,
    ):
        """
        Arguments:
            loss_fn: The classification loss function. If ```None```,
                it defaults to ```torch.nn.CrossEntropyLoss```.
            detach_features: Whether or not to detach the features,
                from which logits are computed.
            f_hook: The hook for computing logits.
        """

        super().__init__(**kwargs)
        self.loss_fn = c_f.default(
            loss_fn, torch.nn.CrossEntropyLoss, {"reduction": "none"}
        )
        self.hook = c_f.default(
            f_hook,
            FeaturesAndLogitsHook,
            {"domains": domains, "detach_features": detach_features},
        )

    def call(self, inputs, losses):
        """"""
        outputs = self.hook(inputs, losses)[0]
        output_losses = {}
        logits = c_f.extract(
            [outputs, inputs],
            c_f.filter(
                self.hook.out_keys, "_logits$", [f"^{d}" for d in self.hook.domains]
            ),
        )
        for i, d in enumerate(self.hook.domains):
            output_losses[self._loss_keys()[i]] = self.loss_fn(
                logits[i], inputs[f"{d}_labels"]
            )
        return outputs, output_losses

    def _loss_keys(self):
        """"""
        return [f"{d}_c_loss" for d in self.hook.domains]


class ClassifierHook(BaseWrapperHook):
    """
    This computes the classification loss and also
    optimizes the models.
    """

    def __init__(
        self,
        opts,
        weighter=None,
        reducer=None,
        loss_fn=None,
        f_hook=None,
        detach_features=False,
        pre=None,
        post=None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        [pre, post] = c_f.many_default([pre, post], [[], []])
        hook = CLossHook(loss_fn, detach_features, f_hook)
        hook = ChainHook(*pre, hook, *post)
        hook = OptimizerHook(hook, opts, weighter, reducer)
        s_hook = SummaryHook({"total_loss": hook})
        self.hook = ChainHook(hook, s_hook)


class FinetunerHook(ClassifierHook):
    """
    This is the same as
    [```ClassifierHook```][pytorch_adapt.hooks.ClassifierHook],
    but it freezes the generator model ("G").
    """

    def __init__(self, **kwargs):
        f_hook = FrozenModelHook(FeaturesHook(detach=True, domains=["src"]), "G")
        f_hook = FeaturesChainHook(f_hook, LogitsHook(domains=["src"]))
        super().__init__(f_hook=f_hook, **kwargs)


# labels are usually int
class BCEWithLogitsConvertLabels(torch.nn.BCEWithLogitsLoss):
    def forward(self, input, target, *args, **kwargs):
        return super().forward(input, target.float(), *args, **kwargs)


class MultiLabelClassifierHook(ClassifierHook):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, loss_fn=BCEWithLogitsConvertLabels(), **kwargs)


class MultiLabelFinetunerHook(FinetunerHook):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, loss_fn=BCEWithLogitsConvertLabels(), **kwargs)



import torch

from ..utils import common_functions as c_f


class ConcatDataset(torch.utils.data.ConcatDataset):
    """
    Exactly the same as ```torch.utils.data.ConcatDataset```
    except with a nice ```__repr__``` function.
    """

    def __repr__(self):
        extra_repr = f"len={str(self.__len__())}"
        return c_f.nice_repr(self, extra_repr, {"datasets": self.datasets})



import albumentations as A
import numpy as np
from albumentations.pytorch.transforms import ToTensorV2

from ..utils import common_functions as c_f
from .constants import IMAGENET_MEAN, IMAGENET_STD


def get_voc_transform(is_training, **kwargs):
    bbox_params = A.BboxParams(
        format="pascal_voc", min_visibility=0.5, label_fields=["class_labels"]
    )
    transform = [A.SmallestMaxSize(224)]
    if is_training:
        transform += [
            A.RandomCrop(height=224, width=224),
            A.HorizontalFlip(p=0.5),
        ]
    else:
        transform += [A.CenterCrop(height=224, width=224)]

    transform += [
        A.Normalize(mean=IMAGENET_MEAN, std=IMAGENET_STD),
        ToTensorV2(),
    ]
    return A.Compose(transform, bbox_params=bbox_params)


class VOCTransformWrapper:
    def __init__(self, transform, labels_to_vec):
        self.transform = transform
        self.labels_to_vec = labels_to_vec

    def __call__(self, image, target):
        objects = target["annotation"]["object"]
        bboxes = [
            [int(y) for y in c_f.extract(x["bndbox"], ["xmin", "ymin", "xmax", "ymax"])]
            for x in objects
        ]
        class_labels = [x["name"] for x in objects]
        x = self.transform(
            image=np.array(image), bboxes=bboxes, class_labels=class_labels
        )
        image = x["image"]
        labels = self.labels_to_vec(x["class_labels"])
        return image, labels

    def __repr__(self):
        return c_f.nice_repr(
            self,
            None,
            {"transform": self.transform, "labels_to_vec": self.labels_to_vec},
        )



from typing import Callable, Dict, List, Tuple

import torch
from pytorch_metric_learning.utils import common_functions as pml_cf

from ..utils import common_functions as c_f


def weight_losses(reduction, weights, scale, loss_dict):
    if "total" in loss_dict:
        raise ValueError("'total' is a reserved key. Use a different loss name")
    c_f.assert_keys_are_present(weights, "weights", loss_dict)
    losses = []
    components = {}
    for k, v in loss_dict.items():
        try:
            loss = v if k not in weights else v * weights[k]
            loss = loss * scale
            losses.append(loss)
            components[k] = loss.item()
        except Exception as e:
            c_f.add_error_message(e, f"\nError occuring with loss key = {k}")
            raise

    total = reduction(losses)
    components["total"] = total.item()
    return total, components


class BaseWeighter:
    """
    Multiplies losses by scalar values, and then reduces them to a single value.
    """

    def __init__(
        self,
        reduction: Callable[[List[torch.Tensor]], torch.Tensor],
        weights: Dict[str, float] = None,
        scale: float = 1,
    ):
        """
        Arguments:
            reduction: A function that takes in a list of losses and returns a single loss value.
            weights: A mapping from loss names to weight values. If ```None```, weights are assumed to be 1.
            scale: A scalar that every loss gets multiplied by.
        """
        self.reduction = reduction
        self.weights = c_f.default(weights, {})
        self.scale = scale
        pml_cf.add_to_recordable_attributes(self, list_of_names=["weights", "scale"])

    def __call__(
        self, loss_dict: Dict[str, torch.Tensor]
    ) -> Tuple[torch.Tensor, Dict[str, float]]:
        """
        Arguments:
            loss_dict: A mapping from loss names to loss values.
        Returns:
            A tuple consisting of

                - the loss that .backward() can be called on

                - a dictionary of floats (detached from the autograd graph)
                        that contains the weighted loss components.
        """
        return weight_losses(self.reduction, self.weights, self.scale, loss_dict)

    def __repr__(self):
        return c_f.nice_repr(self, c_f.extra_repr(self, ["weights", "scale"]), {})



from torchmetrics.functional import accuracy

from .torchmetrics_validator import TorchmetricsValidator


class AccuracyValidator(TorchmetricsValidator):
    """
    Returns accuracy using the
    [torchmetrics accuracy function](https://torchmetrics.readthedocs.io/en/latest/references/functional.html#accuracy-func).

    The required dataset splits are ```["src_val"]```.
    This can be changed using [```key_map```][pytorch_adapt.validators.BaseValidator.__init__].
    """

    @property
    def accuracy_fn(self):
        return accuracy



import numpy as np
import torch

from ..utils import common_functions as c_f
from .base_validator import BaseValidator


def max_normalizer(raw_score_history):
    return raw_score_history / np.abs(np.nanmax(raw_score_history))


def call_val_hook(hook, collected_data, epoch=None):
    kwargs = collected_data
    if hasattr(hook, "required_data"):
        kwargs = c_f.filter_kwargs(collected_data, hook.required_data)
    if isinstance(hook, BaseValidator):
        return hook(**kwargs)
    else:
        return hook(epoch, **kwargs)


##########################
# used by PerClassValidator and TargetKNNValidator


def src_label_fn(x):
    return x["labels"]


def target_label_fn(x):
    return torch.argmax(x["logits"], dim=1)


def default_label_fns(required_data):
    output = {}
    for k in required_data:
        if k.startswith("src"):
            label_fn = src_label_fn
        elif k.startswith("target"):
            label_fn = target_label_fn
        else:
            raise ValueError(f"expected key to start with 'src' or 'target but got {k}")
        output[k] = label_fn
    return output


##########################



from ..containers import KeyEnforcer
from ..hooks import AlignerPlusCHook, RTNHook
from ..inference import rtn_fn
from ..utils import common_functions as c_f
from .base_adapter import BaseGCAdapter
from .utils import with_opt


class Aligner(BaseGCAdapter):
    """
    Wraps [AlignerPlusCHook][pytorch_adapt.hooks.AlignerPlusCHook].

    |Container|Required keys|
    |---|---|
    |models|```["G", "C"]```|
    |optimizers|```["G", "C"]```|
    """

    def init_hook(self, hook_kwargs):
        opts = with_opt(list(self.optimizers.keys()))
        self.hook = self.hook_cls(opts, **hook_kwargs)

    @property
    def hook_cls(self):
        return AlignerPlusCHook


class RTN(Aligner):
    """
    Wraps [RTNHook][pytorch_adapt.hooks.RTNHook].

    |Container|Required keys|
    |---|---|
    |models|```["G", "C", "residual_model"]```|
    |optimizers|```["G", "C", "residual_model"]```|
    |misc|```["feature_combiner"]```|
    """

    def __init__(self, *args, inference_fn=None, **kwargs):
        """
        Arguments:
            inference_fn: Default is [rtn_fn][pytorch_adapt.inference.rtn_fn]
        """
        inference_fn = c_f.default(inference_fn, rtn_fn)
        super().__init__(*args, inference_fn=inference_fn, **kwargs)

    @property
    def hook_cls(self):
        return RTNHook

    def get_key_enforcer(self) -> KeyEnforcer:
        ke = super().get_key_enforcer()
        ke.requirements["models"].append("residual_model")
        ke.requirements["optimizers"].append("residual_model")
        ke.requirements["misc"] = ["feature_combiner"]
        return ke



import torch

from ..layers import SufficientAccuracy
from ..utils import common_functions as c_f
from .base import BaseConditionHook
from .features import DLogitsHook, FeaturesChainHook, FeaturesHook


class StrongDHook(BaseConditionHook):
    """
    Returns ```True``` if the discriminator's
    accuracy is higher than some threshold.
    """

    def __init__(self, threshold: float = 0.6, **kwargs):
        """
        Arguments:
            threshold: The discriminator's accuracy must be higher
                than this threshold for the hook to return ```True```.
        """
        super().__init__(**kwargs)
        self.accuracy_fn = SufficientAccuracy(
            threshold=threshold, to_probs_func=torch.nn.Sigmoid()
        )
        self.hook = FeaturesChainHook(
            FeaturesHook(detach=True), DLogitsHook(detach=True)
        )

    def call(self, inputs, losses):
        """"""
        with torch.no_grad():
            outputs = self.hook(inputs, losses)[0]
            [d_src_logits, d_target_logits] = c_f.extract(
                [outputs, inputs],
                c_f.filter(
                    self.hook.out_keys, "_dlogits_detached$", ["^src", "^target"]
                ),
            )
            [src_domain, target_domain] = c_f.extract(
                inputs, ["src_domain", "target_domain"]
            )
            dlogits = torch.cat([d_src_logits, d_target_logits], dim=0)
            domain_labels = torch.cat([src_domain, target_domain], dim=0)
            return self.accuracy_fn(dlogits, domain_labels)



from abc import ABC
from typing import Callable, Dict

import numpy as np
import torch
from pytorch_metric_learning.utils import common_functions as pml_cf

from ..utils import common_functions as c_f
from .multiple_validators import MultipleValidators


class ScoreHistory(ABC):
    """
    Wraps a [validator][pytorch_adapt.validators.BaseValidator]
    and keeps track of the validator's score history as the epochs progress.
    """

    def __init__(
        self,
        validator,
        normalizer: Callable[[np.ndarray], np.ndarray] = None,
        ignore_epoch: int = None,
    ):
        """
        Arguments:
            normalizer: A function that receives the current unnormalized
                score history, and returns a normalized version of the
                score history. If ```None```, then it defaults to
                no normalization.
            ignore_epoch: This epoch will ignored when determining
                the best scoring epoch. If ```None```, then no epoch will be ignored.
        """
        self.validator = validator
        self.normalizer = c_f.default(normalizer, return_raw)
        self.score_history = np.array([])
        self.raw_score_history = np.array([])
        self.epochs = np.array([], dtype=int)
        self.ignore_epoch = ignore_epoch
        pml_cf.add_to_recordable_attributes(
            self,
            list_of_names=["latest_score", "best_score", "latest_epoch", "best_epoch"],
        )

    def __call__(self, epoch: int, **kwargs: Dict[str, torch.Tensor]) -> float:
        """
        Arguments:
            epoch: The epoch to be scored.
            **kwargs: Keyword arguments that get passed into the wrapped validator's
                [```__call__```][pytorch_adapt.validators.BaseValidator.__call__] method.
        """
        if epoch in self.epochs:
            raise ValueError(f"Epoch {epoch} has already been evaluated")
        score = self.validator(**kwargs)
        sub_scores = None
        if isinstance(score, (list, tuple)):
            score, sub_scores = score
        self.append_to_history_and_normalize(score, epoch)
        if sub_scores:
            return self.latest_score, sub_scores
        return self.latest_score

    def append_to_history_and_normalize(self, score, epoch):
        self.raw_score_history = np.append(self.raw_score_history, score)
        self.epochs = np.append(self.epochs, epoch)
        self.score_history = self.normalizer(self.raw_score_history)

    @property
    def score_history_ignore_epoch(self):
        return remove_ignore_epoch(self.score_history, self.epochs, self.ignore_epoch)

    @property
    def epochs_ignore_epoch(self):
        return remove_ignore_epoch(self.epochs, self.epochs, self.ignore_epoch)

    def has_valid_history(self, ignore_ignore_epoch=True):
        x = (
            self.score_history_ignore_epoch
            if ignore_ignore_epoch
            else self.score_history
        )
        return len(x) > 0 and np.isfinite(x).any()

    @property
    def best_score(self) -> float:
        """
        Returns:
            The best score, ignoring ```self.ignore_epoch```.
                Returns ```None``` if no valid scores are available.
        """
        if self.has_valid_history():
            return self.score_history_ignore_epoch[self.best_idx]

    @property
    def best_epoch(self) -> int:
        """
        Returns:
            The best epoch, ignoring ```self.ignore_epoch```.
                Returns ```None``` if no valid epochs are available.
        """
        if self.has_valid_history():
            return self.epochs_ignore_epoch[self.best_idx]

    @property
    def best_idx(self) -> int:
        """
        Returns:
            The index of the best score in ```self.score_history_ignore_epoch```.
                Returns ```None``` if no valid epochs are available.
        """
        if self.has_valid_history():
            return np.nanargmax(self.score_history_ignore_epoch)

    @property
    def latest_epoch(self) -> int:
        """
        Returns:
            The latest epoch, including ```self.ignore_epoch```.
                Returns ```None``` if no epochs have been scored.
        """
        if self.has_valid_history(False):
            return self.epochs[-1]

    @property
    def latest_score(self) -> float:
        """
        Returns:
            The latest score, including ```self.ignore_epoch```.
                Returns ```None``` if no epochs have been scored.
        """
        if self.has_valid_history(False):
            return self.score_history[-1]

    @property
    def latest_is_best(self) -> bool:
        """
        Returns:
            ```False``` if the latest epoch was not the best
                scoring epoch, or if the latest epoch is ```ignore_epoch```.
                ```True``` otherwise.
        """
        if self.has_valid_history(False):
            return self.best_epoch == self.latest_epoch
        return False

    @property
    def required_data(self):
        return self.validator.required_data

    def __repr__(self):
        return c_f.nice_repr(self, self.extra_repr(), {})

    def extra_repr(self):
        return c_f.extra_repr(
            self, ["validator", "latest_score", "best_score", "best_epoch"]
        )

    def state_dict(self):
        return {k: getattr(self, k) for k in state_dict_keys()}

    def load_state_dict(self, state_dict):
        c_f.assert_state_dict_keys(state_dict, set(state_dict_keys()))
        for k, v in state_dict.items():
            if not isinstance(getattr(self.__class__, k, None), property):
                setattr(self, k, v)


class ScoreHistories(ScoreHistory):
    """
    Like [ScoreHistory][pytorch_adapt.validators.ScoreHistory], but it
    wraps a [MultipleValidators][pytorch_adapt.validators.MultipleValidators] object
    and keeps track of sub-validator scores histories.
    """

    def __init__(self, validator, **kwargs):
        super().__init__(validator=validator, **kwargs)
        if not isinstance(validator, MultipleValidators):
            raise TypeError("validator must be of type MultipleValidators")
        validator.return_sub_scores = True
        self.histories = {k: ScoreHistory(v) for k, v in validator.validators.items()}
        pml_cf.add_to_recordable_attributes(self, list_of_names=["histories"])

    def __call__(self, epoch: int, **kwargs: Dict[str, torch.Tensor]) -> float:
        score, sub_scores = super().__call__(epoch, **kwargs)
        for k, v in self.histories.items():
            v.append_to_history_and_normalize(sub_scores[k], epoch)
        return score

    def extra_repr(self):
        x = super().extra_repr()
        x += f"\n{c_f.extra_repr(self, ['histories'])}"
        return x

    def state_dict(self):
        output = super().state_dict()
        output.update(
            {"histories": {k: v.state_dict() for k, v in self.histories.items()}}
        )
        return output

    def load_state_dict(self, state_dict):
        histories = state_dict.pop("histories")
        super().load_state_dict(state_dict)
        c_f.assert_state_dict_keys(histories, self.histories.keys())
        for k, v in histories.items():
            self.histories[k].load_state_dict(v)


def remove_ignore_epoch(x, epochs, ignore_epoch):
    if ignore_epoch is not None:
        return x[epochs != ignore_epoch]
    return x


def return_raw(raw_score_history):
    return raw_score_history


def state_dict_keys():
    return [
        "best_score",
        "best_epoch",
        "raw_score_history",
        "score_history",
        "epochs",
    ]



from ..containers import KeyEnforcer, MultipleContainers, Optimizers
from ..hooks import (
    ClassifierHook,
    FinetunerHook,
    MultiLabelClassifierHook,
    MultiLabelFinetunerHook,
)
from .base_adapter import BaseGCAdapter
from .utils import default_optimizer_tuple, with_opt


class Classifier(BaseGCAdapter):
    """
    Wraps [ClassifierHook][pytorch_adapt.hooks.ClassifierHook].

    |Container|Required keys|
    |---|---|
    |models|```["G", "C"]```|
    |optimizers|```["G", "C"]```|
    """

    def init_hook(self, hook_kwargs):
        opts = with_opt(list(self.optimizers.keys()))
        self.hook = self.hook_cls(opts=opts, **hook_kwargs)

    @property
    def hook_cls(self):
        return ClassifierHook


class Finetuner(Classifier):
    """
    Wraps [FinetunerHook][pytorch_adapt.hooks.FinetunerHook].

    |Container|Required keys|
    |---|---|
    |models|```["G", "C"]```|
    |optimizers|```["C"]```|
    """

    @property
    def hook_cls(self):
        return FinetunerHook

    def get_default_containers(self) -> MultipleContainers:
        optimizers = Optimizers(default_optimizer_tuple(), keys=["C"])
        return MultipleContainers(optimizers=optimizers)

    def get_key_enforcer(self) -> KeyEnforcer:
        ke = super().get_key_enforcer()
        ke.requirements["optimizers"].remove("G")
        return ke


class MultiLabelClassifier(Classifier):
    @property
    def hook_cls(self):
        return MultiLabelClassifierHook


class MultiLabelFinetuner(Finetuner):
    @property
    def hook_cls(self):
        return MultiLabelFinetunerHook



import os

from torchvision.datasets import VisionDataset, VOCDetection

from .utils import check_length, maybe_download
from .voc_multilabel import process_voc_style_dataset


class Clipart1kMultiLabel(VOCDetection):
    url = "http://www.hal.t.u-tokyo.ac.jp/~inoue/projects/cross_domain_detection/datasets/clipart.zip"
    filename = "clipart1k.zip"
    md5 = "883e2c03eaff39e17ea60d43a3899224"

    def __init__(
        self,
        root: str,
        image_set: str = "train",
        download: bool = False,
        transform=None,
        target_transform=None,
        transforms=None,
    ):
        VisionDataset.__init__(self, root, transforms, transform, target_transform)
        dataset_root = os.path.join(self.root, "clipart1k")

        def rename_fn():
            os.rename(os.path.join(self.root, "clipart"), dataset_root)

        maybe_download(
            process_voc_style_dataset,
            {
                "cls": self,
                "dataset_root": dataset_root,
                "image_set": image_set,
                "download": download,
                "rename_fn": rename_fn,
                "exclude_list": [
                    "681633840"
                ],  # https://github.com/naoto0804/cross-domain-detection/issues/39
            },
        )

        check_length(self, {"train": 499, "test": 500}[image_set])



from .. import utils as f_utils
from .ignite import Ignite


# This is for the case where
# G outputs preds
# C is Identity()
# In other words, the "features" are softmaxed logits
class IgnitePredsAsFeatures(Ignite):
    def __init__(self, *args, **kwargs):
        super().__init__(
            *args,
            val_output_dict_fn=f_utils.create_output_dict_preds_as_features,
            **kwargs
        )



from ..hooks import SymNetsHook
from ..inference import symnets_fn
from ..utils import common_functions as c_f
from .base_adapter import BaseGCAdapter
from .utils import with_opt


class SymNets(BaseGCAdapter):
    """
    Wraps [SymNetsHook][pytorch_adapt.hooks.SymNetsHook].

    |Container|Required keys|
    |---|---|
    |models|```["G", "C"]```|
    |optimizers|```["G", "C"]```|

    The C model must output a list of logits: ```[logits1, logits2]```.
    """

    def __init__(self, *args, inference_fn=None, **kwargs):
        inference_fn = c_f.default(inference_fn, symnets_fn)
        super().__init__(*args, inference_fn=inference_fn, **kwargs)

    def init_hook(self, hook_kwargs):
        self.hook = self.hook_cls(
            g_opts=with_opt(["G"]), c_opts=with_opt(["C"]), **hook_kwargs
        )

    @property
    def hook_cls(self):
        return SymNetsHook



import torch

from ..utils import common_functions as c_f
from .base import BaseWrapperHook
from .features import DLogitsHook, FeaturesChainHook, FeaturesHook, LogitsHook
from .utils import ChainHook


class FeaturesForDomainLossHook(FeaturesChainHook):
    """
    A [```FeaturesChainHook```][pytorch_adapt.hooks.features.FeaturesChainHook]
    that has options specific to
    [```DomainLossHook```][pytorch_adapt.hooks.DomainLossHook].
    """

    def __init__(
        self,
        f_hook=None,
        l_hook=None,
        use_logits=False,
        domains=None,
        detach=False,
        **kwargs,
    ):
        """
        Arguments:
            f_hook: hook for computing features
            l_hook: hook for computing logits. This will be used
                only if ```use_logits``` is ```True```.
            use_logits: If ```True```, the logits hook is executed
                after the features hook.
            domains: the domains for which features will be computed.
            detach: If ```True```, all outputs will be detached
                from the autograd graph.
        """
        hooks = [
            c_f.default(
                f_hook,
                FeaturesHook(detach=detach, domains=domains),
            )
        ]
        if use_logits:
            hooks.append(
                c_f.default(l_hook, LogitsHook(detach=detach, domains=domains))
            )
        super().__init__(*hooks, **kwargs)


class DomainLossHook(BaseWrapperHook):
    """
    Computes the loss of a discriminator's output with
    respect to domain labels.
    """

    def __init__(
        self,
        d_loss_fn=None,
        detach_features=False,
        reverse_labels=False,
        domains=None,
        f_hook=None,
        d_hook=None,
        **kwargs,
    ):
        """
        Arguments:
            d_loss_fn: The loss applied to the discriminator's logits.
                If ```None``` it defaults to
                ```torch.nn.BCEWithLogitsLoss```.
            detach_features: If ```True```, the input to the
                discriminator will be detached first.
            reverse_labels: If ```True```, the ```"src"``` and
                ```"target"``` domain labels will be swapped.
            domains: The domains to apply the loss to.
                If ```None``` it defaults to ```["src", "target"]```.
            f_hook: The hook for computing the input to the discriminator.
            d_hook: The hook for computing the discriminator logits.
        """
        super().__init__(**kwargs)
        self.d_loss_fn = c_f.default(
            d_loss_fn, torch.nn.BCEWithLogitsLoss, {"reduction": "none"}
        )
        self.reverse_labels = reverse_labels
        self.domains = c_f.default(domains, ["src", "target"])
        f_hook = c_f.default(
            f_hook,
            FeaturesForDomainLossHook,
            {"detach": detach_features, "domains": domains},
        )
        d_hook = c_f.default(d_hook, DLogitsHook, {"domains": domains})
        f_out = f_hook.last_hook_out_keys
        d_hook.set_in_keys(f_out)
        self.check_fhook_dhook_keys(f_hook, d_hook, detach_features)
        self.hook = ChainHook(f_hook, d_hook)
        self.in_keys = self.hook.in_keys + ["src_domain", "target_domain"]

    def call(self, inputs, losses):
        losses = {}
        outputs = self.hook(inputs, losses)[0]
        labels = self.extract_domain_labels(inputs)
        for domain_name, labels in labels.items():
            self.logger(f"Computing loss for {domain_name} domain")
            [dlogits] = c_f.extract(
                [outputs, inputs],
                c_f.filter(self.hook.out_keys, "_dlogits$", [f"^{domain_name}"]),
            )
            if dlogits.dim() > 1:
                labels = labels.type(torch.long)
            else:
                labels = labels.type(torch.float)
            loss = self.d_loss_fn(dlogits, labels)
            losses[f"{domain_name}_domain_loss"] = loss
        return outputs, losses

    def extract_domain_labels(self, inputs):
        self.logger("Expecting 'src_domain' and 'target_domain' in inputs")
        [src_domain, target_domain] = c_f.extract(
            inputs, ["src_domain", "target_domain"]
        )
        if self.reverse_labels:
            labels = {"src": target_domain, "target": src_domain}
        else:
            labels = {"src": src_domain, "target": target_domain}
        return {k: v for k, v in labels.items() if k in self.domains}

    def _loss_keys(self):
        return [f"{x}_domain_loss" for x in self.domains]

    def check_fhook_dhook_keys(self, f_hook, d_hook, detach_features):
        if detach_features and len(
            c_f.filter(f_hook.out_keys, "detached$", self.domains)
        ) < len(self.domains):
            error_str = (
                "detach_features is True, but the number of f_hook's detached outputs "
            )
            error_str += "doesn't match the number of domains."
            error_str += f"\nf_hook's outputs: {f_hook.out_keys}"
            error_str += f"\nfdomains: {self.domains}"
            raise ValueError(error_str)
        for name, keys in [("f_hook", f_hook.out_keys), ("d_hook", d_hook.out_keys)]:
            if not all(
                c_f.filter(keys, f"^{self.domains[i]}")
                for i in range(len(self.domains))
            ):
                raise ValueError(
                    f"domains = {self.domains} but d_hook.out_keys = {d_hook.out_keys}"
                )

    def extra_repr(self):
        return c_f.extra_repr(self, ["reverse_labels"])



from .. import utils as f_utils
from .ignite import Ignite


# preds is sigmoid instead of softmax
class IgniteMultiLabelClassification(Ignite):
    def __init__(self, *args, **kwargs):
        super().__init__(
            *args,
            val_output_dict_fn=f_utils.create_output_dict_multilabel_classification,
            **kwargs
        )



from .adabn import AdaBN
from .adda import ADDA
from .aligner import RTN, Aligner
from .base_adapter import BaseAdapter
from .classifier import Classifier, Finetuner, MultiLabelClassifier, MultiLabelFinetuner
from .dann import CDANNE, DANN, DANNE, GVB, GVBE
from .gan import CDAN, CDANE, GAN, GANE, VADA, DomainConfusion
from .mcd import MCD
from .symnets import SymNets



import numpy as np
import torch

from ..utils import common_functions as c_f
from . import utils as val_utils
from .knn_validator import KNNValidator


def get_target_in_ref(query, ref, query_labels, ref_labels):
    data = []
    for L in torch.unique(query_labels):
        mask = query_labels == L
        curr_query = query[mask]
        curr_ref = torch.cat([ref, query[~mask]], dim=0)
        curr_query_labels = query_labels[mask]
        curr_ref_labels = torch.cat([ref_labels, query_labels[~mask]], dim=0)
        if len(curr_query) != len(curr_query_labels) or len(curr_ref) != len(
            curr_ref_labels
        ):
            raise ValueError("lengths should match")
        data.append((curr_query, curr_ref, curr_query_labels, curr_ref_labels))
    return data


class TargetKNNValidator(KNNValidator):
    def __init__(
        self, add_target_to_ref=False, src_label_fn=None, target_label_fn=None, **kwargs
    ):
        super().__init__(**kwargs)
        self.add_target_to_ref = add_target_to_ref
        self.src_label_fn = c_f.default(src_label_fn, val_utils.src_label_fn)
        self.target_label_fn = c_f.default(target_label_fn, val_utils.target_label_fn)

    def compute_score(self, src_train, target_train):
        query = target_train[self.layer]
        query_labels = self.target_label_fn(target_train)
        ref = src_train[self.layer]
        ref_labels = self.src_label_fn(src_train)

        if self.add_target_to_ref:
            data = get_target_in_ref(query, ref, query_labels, ref_labels)
        else:
            data = [(query, ref, query_labels, ref_labels)]

        scores = []
        for d in data:
            accuracies = self.acc_fn.get_accuracy(
                *d, embeddings_come_from_same_source=False
            )
            scores.append(accuracies[self.metric])
        return np.mean(scores)



import torch
import torch.nn.functional as F
from pytorch_metric_learning.distances import BatchedDistance, CosineSimilarity

from ..layers import EntropyLoss
from ..utils.common_functions import mask_out_self
from .base_validator import BaseValidator


def get_iter_fn(all_entropies, entropy_fn, T):
    def fn(sim_mat, s, *_):
        sim_mat = mask_out_self(sim_mat, s)
        sim_mat = F.softmax(sim_mat / T, dim=1)
        all_entropies.append(entropy_fn(sim_mat))

    return fn


# https://arxiv.org/pdf/2108.10860.pdf
class SNDValidator(BaseValidator):
    """
    Implementation of
    [Tune it the Right Way: Unsupervised Validation of Domain Adaptation via Soft Neighborhood Density](https://arxiv.org/abs/2108.10860)
    """

    def __init__(self, layer="preds", T=0.05, batch_size=1024, **kwargs):
        super().__init__(**kwargs)
        self.layer = layer
        self.T = T
        self.entropy_fn = EntropyLoss(after_softmax=True, return_mean=False)
        self.dist_fn = BatchedDistance(CosineSimilarity(), batch_size=batch_size)

    def compute_score(self, target_train):
        features = target_train[self.layer]
        # all_entropies is modified via self.iter_fn
        all_entropies = []
        self.dist_fn.iter_fn = get_iter_fn(all_entropies, self.entropy_fn, self.T)
        self.dist_fn(features)
        all_entropies = torch.cat(all_entropies, dim=0)
        if len(all_entropies) != len(features):
            raise ValueError("all_entropies should have same length as input features")
        return torch.mean(all_entropies).item()



from torchmetrics.functional import average_precision

from .torchmetrics_validator import TorchmetricsValidator


class APValidator(TorchmetricsValidator):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.layer != "preds":
            raise ValueError("layer must be 'preds'")
        if "num_classes" not in self.torchmetric_kwargs:
            raise ValueError("'num_classes' must be provided in torchmetric_kwargs")

    @property
    def accuracy_fn(self):
        return average_precision



from .base_container import BaseContainer


class Misc(BaseContainer):
    """
    This is used for modules that don't
    have any optimizable parameters, and other
    miscellaneous objects.
    """

    pass



import os

from torchvision import datasets as torch_datasets

from .base_dataset import BaseDataset, BaseDownloadableDataset
from .utils import check_img_paths, check_length, check_train


class OfficeHomeFull(BaseDataset):
    """
    A dataset consisting of 65 classes in 4 domains:
    art, clipart, product, and real.
    """

    def __init__(self, root: str, domain: str, transform):
        """
        Arguments:
            root: The dataset must be located at ```<root>/officehome```
            domain: One of ```"art", "clipart", "product", "real"```.
            transform: The image transform applied to each sample.
        """
        super().__init__(domain=domain)
        self.transform = transform
        self.dataset = torch_datasets.ImageFolder(
            os.path.join(root, "officehome", domain),
            transform=self.transform,
        )
        check_length(
            self, {"art": 2427, "clipart": 4365, "product": 4439, "real": 4357}[domain]
        )

    def __len__(self):
        return len(self.dataset)

    def __getitem__(self, idx):
        return self.dataset[idx]


class OfficeHome(BaseDownloadableDataset):
    """
    A custom train/test split of [OfficeHomeFull][pytorch_adapt.datasets.OfficeHomeFull].

    Extends [BaseDownloadableDataset][pytorch_adapt.datasets.BaseDownloadableDataset],
    so the dataset can be downloaded by setting ```download=True``` when
    initializing.
    """

    url = "https://cornell.box.com/shared/static/xwsbubtcr8flqfuds5f6okqbr3z0w82t"
    filename = "officehome_resized.tar.gz"
    md5 = "52d6039512434aa561d66de9c10828c3"

    def __init__(self, root: str, domain: str, train: bool, transform=None, **kwargs):
        """
        Arguments:
            root: The dataset must be located at ```<root>/officehome```
            domain: One of ```"art", "clipart", "product", "real"```.
            train: Whether or not to use the training set.
            transform: The image transform applied to each sample.
        """
        self.train = check_train(train)
        super().__init__(root=root, domain=domain, **kwargs)
        self.transform = transform

    def set_paths_and_labels(self, root):
        name = "train" if self.train else "test"
        labels_file = os.path.join(root, "officehome", f"{self.domain}_{name}.txt")
        img_dir = os.path.join(root, "officehome")

        with open(labels_file) as f:
            content = [line.rstrip().split(" ") for line in f]
        self.img_paths = [os.path.join(img_dir, x[0]) for x in content]
        check_img_paths(img_dir, self.img_paths, self.domain)
        check_length(
            self,
            {
                "art": {"train": 1941, "test": 486}[name],
                "clipart": {"train": 3492, "test": 873}[name],
                "product": {"train": 3551, "test": 888}[name],
                "real": {"train": 3485, "test": 872}[name],
            }[self.domain],
        )
        self.labels = [int(x[1]) for x in content]



from ..containers import KeyEnforcer, MultipleContainers
from ..hooks import AdaBNHook
from ..inference import adabn_fn
from ..utils import common_functions as c_f
from .base_adapter import BaseAdapter


class AdaBN(BaseAdapter):
    """
    Wraps [AdaBNHook][pytorch_adapt.hooks.AdaBNHook].

    |Container|Required keys|
    |---|---|
    |models|```["G", "C"]```|
    """

    def __init__(self, *args, inference_fn=None, **kwargs):
        """
        Arguments:
            inference_fn: Default is [adabn_fn][pytorch_adapt.inference.adabn_fn]
        """
        inference_fn = c_f.default(inference_fn, adabn_fn)
        super().__init__(*args, inference_fn=inference_fn, **kwargs)

    def init_hook(self, hook_kwargs):
        self.hook = self.hook_cls(**hook_kwargs)

    @property
    def hook_cls(self):
        return AdaBNHook

    def get_key_enforcer(self) -> KeyEnforcer:
        return KeyEnforcer(models=["G", "C"], optimizers=[])

    def get_default_containers(self) -> MultipleContainers:
        return MultipleContainers()



from .accuracy_validator import AccuracyValidator
from .ap_validator import APValidator
from .base_validator import BaseValidator
from .bnm_validator import BNMValidator
from .class_cluster_validator import ClassClusterValidator
from .deep_embedded_validator import DeepEmbeddedValidator
from .diversity_validator import DiversityValidator
from .entropy_validator import EntropyValidator
from .error_validator import ErrorValidator
from .im_validator import IMValidator
from .ist_validator import ISTValidator
from .knn_validator import KNNValidator
from .mmd_validator import MMDValidator
from .multiple_validators import MultipleValidators
from .nearest_source_validator import NearestSourceL2Validator, NearestSourceValidator
from .per_class_validator import PerClassValidator
from .score_history import ScoreHistories, ScoreHistory
from .snd_validator import SNDValidator
from .target_knn_validator import TargetKNNValidator



from collections import defaultdict

import numpy as np
import torch

from ..utils import common_functions as c_f
from .base_validator import BaseValidator
from .utils import default_label_fns


def check_keys(src_key, target_key):
    if not src_key.startswith("src") or not target_key.startswith("target"):
        raise ValueError(
            f"Expected keys to start with 'src' and 'target', but got {src_key}, {target_key}"
        )


def get_common_labels(kwargs, label_fns):
    labels = {}
    label_sets = []
    for split_name, v1 in kwargs.items():
        curr_labels = label_fns[split_name](v1)
        labels[split_name] = curr_labels
        label_sets.append(set(torch.unique(curr_labels).tolist()))
    common_labels = set.intersection(*label_sets)
    return labels, common_labels


def create_new_kwargs(kwargs, labels, common_labels):
    new_kwargs = []
    for x in common_labels:
        new_kwargs.append(defaultdict(dict))
        for split_name, v1 in kwargs.items():
            curr_labels = labels[split_name]
            for feature_name, v2 in v1.items():
                new_kwargs[-1][split_name][feature_name] = v2[curr_labels == x]
    return new_kwargs


class PerClassValidator(BaseValidator):
    def __init__(self, validator, label_fns=None, **kwargs):
        super().__init__(**kwargs)
        self.validator = validator
        self.label_fns = c_f.default(label_fns, default_label_fns, [self.required_data])
        if self.key_map != {}:
            raise ValueError("key_map should be passed to the wrapped validator")

    def _required_data(self):
        return self.validator.required_data

    def compute_score(self, **kwargs):
        labels, common_labels = get_common_labels(kwargs, self.label_fns)
        new_kwargs = create_new_kwargs(kwargs, labels, common_labels)
        scores = [self.validator(**nk) for nk in new_kwargs]
        return np.mean(scores)

    def extra_repr(self):
        x = super().extra_repr()
        x += f"\n{c_f.extra_repr(self, ['validator'])}"
        return x



import numpy as np
import torch
from sklearn.cluster import KMeans
from sklearn.decomposition import PCA
from sklearn.metrics import adjusted_mutual_info_score

from ..utils import common_functions as c_f
from . import utils as val_utils
from .base_validator import BaseValidator


def check_centroid_init(centroid_init):
    if centroid_init not in ["label_centers", None]:
        raise ValueError("centroid_init should be 'label_centers' or None")


class ClassClusterValidator(BaseValidator):
    def __init__(
        self,
        layer="features",
        src_label_fn=None,
        target_label_fn=None,
        with_src=False,
        src_for_pca=False,
        pca_size=None,
        centroid_init=None,
        score_fn=None,
        score_fn_type="labels",
        feat_normalizer=None,
        **kwargs
    ):
        super().__init__(**kwargs)
        self.layer = layer
        self.src_label_fn = c_f.default(src_label_fn, val_utils.src_label_fn)
        self.target_label_fn = c_f.default(target_label_fn, val_utils.target_label_fn)
        self.with_src = with_src
        self.src_for_pca = src_for_pca
        self.pca_size = pca_size
        check_centroid_init(centroid_init)
        self.centroid_init = centroid_init
        self.score_fn = c_f.default(score_fn, adjusted_mutual_info_score)
        self.score_fn_type = score_fn_type
        self.feat_normalizer = feat_normalizer

    def _required_data(self):
        x = ["target_train"]
        if self.requires_src:
            x.append("src_train")
        return x

    @property
    def requires_src(self):
        return self.with_src or self.src_for_pca

    def compute_score(self, target_train, src_train=None):
        feats = target_train[self.layer]
        labels = self.target_label_fn(target_train)
        num_classes = target_train["logits"].shape[1]

        src_feats, src_labels = None, None
        if self.requires_src:
            src_feats = src_train[self.layer]
        if self.with_src:
            src_labels = src_train["labels"]

        return get_clustering_performance(
            feats,
            labels,
            num_classes,
            self.score_fn,
            self.score_fn_type,
            src_feats=src_feats,
            src_labels=src_labels,
            pca_size=self.pca_size,
            centroid_init=self.centroid_init,
            feat_normalizer=self.feat_normalizer,
        )


# copied from https://github.com/lr94/abas/blob/master/model_selection.py
def get_centroids(data, labels, num_classes):
    centroids = np.zeros((num_classes, data.shape[1]))
    for cid in range(num_classes):
        # Since we are using pseudolabels to compute centroids, some classes might not have instances according to the
        # pseudolabels assigned by the current model. In that case .mean() would return NaN causing KMeans to fail.
        # We set to 0 the missing centroids
        if (labels == cid).any():
            centroids[cid] = data[labels == cid].mean(0)

    return centroids


# adapted from https://github.com/lr94/abas/blob/master/model_selection.py
def get_clustering_performance(
    feats,
    labels,
    num_classes,
    score_fn,
    score_fn_type,
    src_feats=None,
    src_labels=None,
    pca_size=None,
    centroid_init=None,
    feat_normalizer=None,
):
    num_target_feats = feats.shape[0]

    if src_feats is not None:
        feats = torch.cat((feats, src_feats), dim=0)

    if feat_normalizer:
        feats = feat_normalizer(feats)

    feats = feats.cpu().numpy()

    if pca_size is not None:
        pca = PCA(pca_size)
        feats = pca.fit_transform(feats)

    # no labels means src_feats is just for pca
    if src_labels is None:
        feats = feats[:num_target_feats]
    else:
        labels = torch.cat((labels, src_labels), dim=0)

    labels = labels.cpu().numpy()

    if centroid_init == "label_centers":
        centroids = get_centroids(feats, labels, num_classes)
        clustering = KMeans(n_clusters=num_classes, init=centroids, n_init=1)
    elif centroid_init is None:
        clustering = KMeans(n_clusters=num_classes)

    clustering.fit(feats)
    clabels = clustering.labels_

    if score_fn_type == "labels":
        score = score_fn(labels, clabels)
    elif score_fn_type == "features":
        score = score_fn(feats, clabels)

    return score



from typing import Any, Dict

from torch.utils.data import Dataset

from .domain_dataset import DomainDataset


class SourceDataset(DomainDataset):
    """
    Wrap your source dataset with this. Your source dataset's
    ```__getitem__``` function should return a tuple of ```(data, label)```.
    """

    def __init__(self, dataset: Dataset, domain: int = 0):
        """
        Arguments:
            dataset: The dataset to wrap
            domain: An integer representing the domain.
        """
        super().__init__(dataset, domain)

    def __getitem__(self, idx: int) -> Dict[str, Any]:
        """
        Returns:
            A dictionary with keys

                - "src_imgs" (the data)

                - "src_domain" (the integer representing the domain)

                - "src_labels" (the class label)

                - "src_sample_idx" (idx)
        """

        img, src_labels = self.dataset[idx]
        return {
            "src_imgs": img,
            "src_domain": self.domain,
            "src_labels": src_labels,
            "src_sample_idx": idx,
        }



from .adabn import AdaBNHook
from .adda import ADDAHook
from .aligners import (
    AlignerHook,
    AlignerPlusCHook,
    FeaturesLogitsAlignerHook,
    JointAlignerHook,
    ManyAlignerHook,
)
from .atdoc import ATDOCHook
from .cdan import CDANDomainHookD, CDANDomainHookG, CDANEHook, CDANHook, GANEHook
from .classification import (
    ClassifierHook,
    CLossHook,
    FinetunerHook,
    MultiLabelClassifierHook,
    MultiLabelFinetunerHook,
    SoftmaxHook,
)
from .conditions import StrongDHook
from .dann import (
    CDANNEHook,
    DANNEHook,
    DANNHook,
    DANNLogitsHook,
    DANNSoftmaxLogitsHook,
    GradientReversalHook,
    GradientReversalLocallyHook,
    SoftmaxGradientReversalHook,
    SoftmaxGradientReversalLocallyHook,
)
from .domain import DomainLossHook, FeaturesForDomainLossHook
from .domain_confusion import DomainConfusionHook
from .features import (
    CombinedFeaturesHook,
    DLogitsHook,
    FeaturesAndLogitsHook,
    FeaturesHook,
    FeaturesWithGradAndDetachedHook,
    FrozenModelHook,
    LogitsHook,
)
from .gan import GANHook
from .gvb import GVBEHook, GVBGANHook, GVBHook
from .itl import ISTLossHook
from .losses import (
    AFNHook,
    BaseLossHook,
    BNMHook,
    BSPHook,
    MCCHook,
    TargetDiversityHook,
    TargetEntropyHook,
)
from .mcd import MCDHook, MCDLossHook, MultipleCLossHook
from .reducers import EntropyReducer, MeanReducer, MultipleReducers
from .rtn import ResidualHook, RTNAlignerHook, RTNHook, RTNLogitsHook
from .symnets import (
    SymNetsCategoryLossHook,
    SymNetsCHook,
    SymNetsDomainLossHook,
    SymNetsEntropyHook,
    SymNetsGHook,
    SymNetsHook,
)
from .utils import (
    ApplyToListHook,
    AssertHook,
    ChainHook,
    EmptyHook,
    FalseHook,
    ParallelHook,
    RepeatHook,
    TrueHook,
)
from .vada import VADAHook, VATHook, VATPlusEntropyHook
from .validate import validate_hook



import os

import torch
from torchvision.datasets import VOCDetection
from torchvision.datasets.utils import download_and_extract_archive

from ..utils import common_functions as c_f
from .utils import check_length, maybe_download

NUM_CLASSES = 20
CLASSNAMES = [
    "aeroplane",
    "bicycle",
    "bird",
    "boat",
    "bottle",
    "bus",
    "car",
    "cat",
    "chair",
    "cow",
    "diningtable",
    "dog",
    "horse",
    "motorbike",
    "person",
    "pottedplant",
    "sheep",
    "sofa",
    "train",
    "tvmonitor",
]
CLASSNAME_TO_IDX = {x: i for i, x in enumerate(CLASSNAMES)}


# modified from https://github.com/pytorch/vision/blob/main/torchvision/datasets/voc.py
def process_voc_style_dataset(
    cls, dataset_root, image_set, download, rename_fn=None, exclude_list=None
):
    if download:
        download_and_extract_archive(
            cls.url, cls.root, filename=cls.filename, md5=cls.md5
        )
        if rename_fn:
            rename_fn()

    splits_dir = os.path.join(dataset_root, "ImageSets", cls._SPLITS_DIR)
    split_f = os.path.join(splits_dir, image_set.rstrip("\n") + ".txt")
    with open(os.path.join(split_f)) as f:
        file_names = [x.strip() for x in f.readlines()]

    exclude_list = c_f.default(exclude_list, [])
    file_names = [x for x in file_names if x not in exclude_list]
    image_dir = os.path.join(dataset_root, "JPEGImages")
    cls.images = [os.path.join(image_dir, x + ".jpg") for x in file_names]

    target_dir = os.path.join(dataset_root, cls._TARGET_DIR)
    cls.targets = [
        os.path.join(target_dir, x + cls._TARGET_FILE_EXT) for x in file_names
    ]

    assert len(cls.images) == len(cls.targets)


def get_labels_as_vector(class_names):
    classes = [CLASSNAME_TO_IDX[x] for x in class_names]
    label = torch.zeros(NUM_CLASSES, dtype=int)
    label[classes] = 1
    return label


class VOCMultiLabel(VOCDetection):
    def __init__(self, **kwargs):
        maybe_download(super().__init__, kwargs)

        check_length(
            self,
            {
                "2007": {"train": 2501, "trainval": 5011, "val": 2510, "test": 4952},
                "2008": {"train": 2111, "trainval": 4332, "val": 2221},
                "2009": {"train": 3473, "trainval": 7054, "val": 3581},
                "2010": {"train": 4998, "trainval": 10103, "val": 5105},
                "2011": {"train": 5717, "trainval": 11540, "val": 5823},
                "2012": {"train": 5717, "trainval": 11540, "val": 5823},
            }[self.year][self.image_set],
        )



from .base_container import BaseContainer


class Models(BaseContainer):
    """
    A container with some functions specific to models
    that have optimizable parameters.
    """

    def train(self):
        """
        Sets all models to train mode.
        """
        for v in self.values():
            v.train()

    def eval(self):
        """
        Sets all models to eval mode.
        """
        for v in self.values():
            v.eval()

    def zero_grad(self):
        """
        Zeros the gradients in all models.
        """
        for v in self.values():
            v.zero_grad()

    def to(self, device):
        """
        Moves all models to ```device```.
        """
        for v in self.values():
            v.to(device)



from ..layers import DiversityLoss
from .simple_loss_validator import SimpleLossValidator


class DiversityValidator(SimpleLossValidator):
    """
    Returns the negative of the
    [diversity][pytorch_adapt.layers.diversity_loss.DiversityLoss]
    of all logits.
    """

    @property
    def loss_fn(self):
        return DiversityLoss(after_softmax=self.layer == "preds")



from typing import Callable

import torch

from ..utils import common_functions as c_f
from .normalizers import NoNormalizer


class ConfidenceWeights(torch.nn.Module):
    """
    Returns the max value along each row of the input,
    followed by an optional normalization function.
    The output of this can be used to weight
    classification losses by the "confidence" of the predictions.
    """

    def __init__(self, normalizer: Callable[[torch.Tensor], torch.Tensor] = None):
        """
        Arguments:
            normalizer: A callable for normalizing
                (e.g. min-max normalization) the weights.
                If ```None```, then no normalization is used.
        """

        super().__init__()
        self.normalizer = c_f.default(normalizer, NoNormalizer())

    def forward(self, preds):
        """"""
        return self.normalizer(torch.max(preds, dim=1)[0])



from ..layers import EntropyWeights, MaxNormalizer
from ..utils import common_functions as c_f
from .base import BaseWrapperHook
from .classification import SoftmaxLocallyHook
from .domain import DomainLossHook
from .features import CombinedFeaturesHook, FeaturesAndLogitsHook
from .gan import GANHook
from .reducers import EntropyReducer, MeanReducer
from .utils import ChainHook


def cdan_key_map(fc_hook):
    strs = c_f.filter(fc_hook.out_keys, "", ["^src", "^target"])
    return {k: v for k, v in zip(strs, ["src_imgs_features", "target_imgs_features"])}


def get_entropy_reducer(apply_to, detach_weights):
    return EntropyReducer(
        apply_to=apply_to,
        default_reducer=MeanReducer(),
        entropy_weights_fn=EntropyWeights(normalizer=MaxNormalizer(detach=True)),
        detach_weights=detach_weights,
    )


def get_entropy_reducers_for_gan(detach_entropy_reducer):
    d_reducer = get_entropy_reducer(
        ["d_src_domain_loss", "d_target_domain_loss"],
        detach_weights=True,
    )
    g_reducer = get_entropy_reducer(
        ["g_src_domain_loss", "g_target_domain_loss"],
        detach_weights=detach_entropy_reducer,
    )
    return d_reducer, g_reducer


def get_cdan_features_hooks(softmax):
    f_hook = FeaturesAndLogitsHook()
    fc_hook = CombinedFeaturesHook()
    key_map = cdan_key_map(fc_hook)
    if softmax:
        strs = c_f.filter(f_hook.out_keys, "_features_logits$", ["^src", "^target"])
        fc_hook = SoftmaxLocallyHook(strs, fc_hook)
    return f_hook, fc_hook, key_map


class CDANDomainHook(BaseWrapperHook):
    def __init__(
        self, loss_prefix, detach_features, reverse_labels, softmax=True, **kwargs
    ):
        super().__init__(**kwargs)
        f_hook, fc_hook, key_map = get_cdan_features_hooks(softmax)
        d_hook = DomainLossHook(
            loss_prefix=loss_prefix,
            detach_features=detach_features,
            reverse_labels=reverse_labels,
            key_map=key_map,
        )
        self.hook = ChainHook(f_hook, fc_hook, d_hook)


class CDANDomainHookD(CDANDomainHook):
    def __init__(self, **kwargs):
        super().__init__(
            loss_prefix="d_", detach_features=True, reverse_labels=False, **kwargs
        )


class CDANDomainHookG(CDANDomainHook):
    def __init__(self, **kwargs):
        super().__init__(
            loss_prefix="g_", detach_features=False, reverse_labels=True, **kwargs
        )


class CDANHook(GANHook):
    """
    Implementation of
    [Conditional Adversarial Domain Adaptation](https://arxiv.org/abs/1705.10667)
    """

    def __init__(self, softmax=True, **kwargs):
        super().__init__(
            disc_hook=CDANDomainHookD(softmax=softmax),
            gen_hook=CDANDomainHookG(softmax=softmax),
            **kwargs
        )


class CDANEHook(CDANHook):
    def __init__(self, detach_entropy_reducer=True, **kwargs):
        d_reducer, g_reducer = get_entropy_reducers_for_gan(detach_entropy_reducer)
        super().__init__(d_reducer=d_reducer, g_reducer=g_reducer, **kwargs)


class GANEHook(GANHook):
    def __init__(self, detach_entropy_reducer=True, **kwargs):
        d_reducer, g_reducer = get_entropy_reducers_for_gan(detach_entropy_reducer)
        super().__init__(d_reducer=d_reducer, g_reducer=g_reducer, **kwargs)



from abc import ABC, abstractmethod
from typing import Any, Dict, Tuple

import torch

from ..containers import (
    KeyEnforcer,
    LRSchedulers,
    Misc,
    Models,
    MultipleContainers,
    Optimizers,
)
from ..inference import default_fn
from ..utils import common_functions as c_f
from .utils import default_optimizer_tuple, with_opt


class BaseAdapter(ABC):
    """
    Parent class of all adapters.
    """

    def __init__(
        self,
        models: Models = None,
        optimizers: Optimizers = None,
        lr_schedulers: LRSchedulers = None,
        misc: Misc = None,
        default_containers: MultipleContainers = None,
        key_enforcer: KeyEnforcer = None,
        inference_fn=None,
        before_training_starts=None,
        hook_kwargs: Dict[str, Any] = None,
    ):
        """
        Arguments:
            models: A [```Models```][pytorch_adapt.containers.Models] container.
                The models will be passed to the wrapped hook at each
                training iteration.
            optimizers: An [```Optimizers```][pytorch_adapt.containers.Optimizers] container.
                The optimizers will be passed to the wrapped hook at each
                training iteration.
            lr_schedulers: An [```LRSchedulers```][pytorch_adapt.containers.LRSchedulers] container.
                The lr schedulers are called automatically by the
                [```framework```](../frameworks/index.md) that wrap this adapter.
            misc: A [```Misc```][pytorch_adapt.containers.Misc] container for models
                that don't require optimizers, and other miscellaneous objects.
                These are passed into the wrapped hook at each training iteration.
            default_containers: The default set of containers to use, wrapped in a
                [```MultipleContainers```][pytorch_adapt.containers.MultipleContainers] object.
                If ```None``` then the default containers are defined in
                [```self.get_default_containers```][pytorch_adapt.adapters.BaseAdapter.get_default_containers]
            key_enforcer: A [```KeyEnforcer```][pytorch_adapt.containers.KeyEnforcer] object.
                If ```None```, then [```self.get_key_enforcer```][pytorch_adapt.adapters.BaseAdapter.get_key_enforcer]
                is used.
            inference_fn: A function that takes in:

                - ```x```: the input to the model
                - ```domain```: an integer representing the domain of the data
                - ```models```: a dictionary of models, i.e. ```self.models```
                - ```misc```: a dictionary of misc objects, i.e. ```self.misc```

            before_training_starts: A function that takes in this adapter and returns another
                function that is optionally called by a framework wrapper before training starts.
            hook_kwargs: A dictionary of keyword arguments that will be
                passed into the wrapped hook during initialization.
        """
        containers = c_f.default(default_containers, self.get_default_containers, {})
        self.key_enforcer = c_f.default(key_enforcer, self.get_key_enforcer, {})
        self.before_training_starts = c_f.class_default(
            self, before_training_starts, self.before_training_starts_default
        )

        containers.merge(
            models=models,
            optimizers=optimizers,
            lr_schedulers=lr_schedulers,
            misc=misc,
        )

        hook_kwargs = c_f.default(hook_kwargs, {})
        self.init_containers_and_check_keys(containers)
        self.init_hook(hook_kwargs)
        self.inference_fn = c_f.default(inference_fn, default_fn)

    def training_step(
        self, batch: Dict[str, Any], **kwargs
    ) -> Dict[str, Dict[str, float]]:
        """
        Calls the wrapped hook at each iteration during training.
        Arguments:
            batch: A dictionary containing training data.
            **kwargs: Any other data that will be passed into the hook.
        Returns:
            A two-level dictionary

                - the outer level is associated with a particular optimization step
                    (relevant for GAN architectures)

                - the inner level contains the loss components.
        """
        combined = c_f.assert_dicts_are_disjoint(
            self.models, self.misc, with_opt(self.optimizers), batch, kwargs
        )
        _, losses = self.hook(combined)
        return losses

    def inference(
        self, x: torch.Tensor, domain: int = None
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Arguments:
            x: The input to the model
            domain: An optional integer indicating the domain.

        Returns:
            Features and logits
        """
        return self.inference_fn(
            x=x,
            domain=domain,
            models=self.models,
            misc=self.misc,
        )

    def get_default_containers(self) -> MultipleContainers:
        """
        Returns:
            The default set of containers. Consists of an
                [Optimizers][pytorch_adapt.containers.Optimizers]
                container using the [default][pytorch_adapt.adapters.utils.default_optimizer_tuple]
                of Adam with lr 0.0001.
        """
        optimizers = Optimizers(default_optimizer_tuple())
        return MultipleContainers(optimizers=optimizers)

    @abstractmethod
    def get_key_enforcer(self) -> KeyEnforcer:
        """
        Returns:
            The default KeyEnforcer.
        """
        pass

    @abstractmethod
    def init_hook(self):
        """
        ```self.hook``` is initialized here.
        """
        pass

    @property
    @abstractmethod
    def hook_cls(self):
        pass

    def init_containers_and_check_keys(self, containers):
        """
        Called in ```__init__``` before
        [```init_hook```][pytorch_adapt.adapters.BaseAdapter.init_hook].
        """
        containers.create()
        self.key_enforcer.check(containers)
        for k, v in containers.items():
            setattr(self, k, v)

    def before_training_starts_default(self, framework):
        c_f.LOGGER.debug(f"models\n{self.models}")
        c_f.LOGGER.debug(f"optimizers\n{self.optimizers}")
        c_f.LOGGER.debug(f"lr_schedulers\n{self.lr_schedulers}")
        c_f.LOGGER.debug(f"misc\n{self.misc}")
        c_f.LOGGER.debug(f"hook\n{self.hook}")


class BaseGCDAdapter(BaseAdapter):
    """
    Base class for adapters that use a Generator, Classifier, and Discriminator.
    """

    def get_key_enforcer(self) -> KeyEnforcer:
        return KeyEnforcer(
            models=["G", "C", "D"],
            optimizers=["G", "C", "D"],
        )


class BaseGCAdapter(BaseAdapter):
    """
    Base class for adapters that use a Generator and Classifier.
    """

    def get_key_enforcer(self) -> KeyEnforcer:
        return KeyEnforcer(
            models=["G", "C"],
            optimizers=["G", "C"],
        )



import numpy as np
import torch
from pytorch_metric_learning.utils import common_functions as pml_cf

from ..utils import common_functions as c_f


def split_half(x, dim):
    d = x.shape[dim] // 2
    return torch.split(x, d, dim=dim)


def num_elements_minus_diag(x):
    n = x.shape[0]
    return n * (n - 1)


def get_kernel_scales(low=-8, high=8, num_kernels=33, base=2.0):
    return torch.from_numpy(np.logspace(low, high, num=num_kernels, base=base))


def _mmd_dist_mats(x, y, dist_func, bandwidth=None):
    xx = dist_func(x, x)
    yy = dist_func(y, y)
    zz = dist_func(x, y)

    with torch.no_grad():
        # https://arxiv.org/pdf/1409.6041.pdf
        # https://arxiv.org/pdf/1707.07269.pdf
        denom = (
            torch.median(xx)
            if bandwidth is None
            else torch.tensor([bandwidth], dtype=xx.dtype, device=xx.device)
        )
        scale = -1.0 / denom

    return xx, yy, zz, scale


def get_mmd_dist_mats(x, y, dist_func, bandwidth):
    if c_f.is_list_or_tuple(x):
        xx, yy, zz, scale = [], [], [], []
        for i in range(len(x)):
            _xx, _yy, _zz, _scale = _mmd_dist_mats(x[i], y[i], dist_func, bandwidth)
            xx.append(_xx)
            yy.append(_yy)
            zz.append(_zz)
            scale.append(_scale)
        return xx, yy, zz, scale
    else:
        return _mmd_dist_mats(x, y, dist_func, bandwidth)


def get_default_kernel_weights(scale):
    if torch.is_tensor(scale) and torch.numel(scale) > 1:
        return torch.ones_like(scale) / len(scale)
    else:
        return 1


def _mmd_quadratic(x, scale, weights):
    return torch.sum(torch.exp(x.unsqueeze(2) * scale) * weights, dim=2)


def get_mmd_quadratic(xx, yy, zz, scale, weights=None):
    # https://jmlr.csail.mit.edu/papers/volume13/gretton12a/gretton12a.pdf
    # https://arxiv.org/pdf/1502.02791.pdf
    is_joint_mmd = c_f.is_list_or_tuple(xx)

    if is_joint_mmd:
        xx_prod, yy_prod, zz_prod = 1, 1, 1
        for i in range(len(xx)):
            curr_weights = c_f.default(weights, get_default_kernel_weights(scale[i]))
            xx_prod *= _mmd_quadratic(xx[i], scale[i], curr_weights)
            yy_prod *= _mmd_quadratic(yy[i], scale[i], curr_weights)
            zz_prod *= _mmd_quadratic(zz[i], scale[i], curr_weights)
        xx_prod.fill_diagonal_(0)
        yy_prod.fill_diagonal_(0)
        xx, yy, zz = xx_prod, yy_prod, zz_prod
    else:
        weights = c_f.default(weights, get_default_kernel_weights(scale))
        xx = _mmd_quadratic(xx, scale, weights).fill_diagonal_(0)
        yy = _mmd_quadratic(yy, scale, weights).fill_diagonal_(0)
        zz = _mmd_quadratic(zz, scale, weights)

    xx_scaler = 1.0 / num_elements_minus_diag(xx)
    yy_scaler = 1.0 / num_elements_minus_diag(yy)
    return xx_scaler * torch.sum(xx) + yy_scaler * torch.sum(yy) - 2 * torch.mean(zz)


def _mmd_linear(x, i, j, scale, weights):
    return torch.sum(torch.exp(x[i, j] * scale) * weights, dim=0)


def _mmd_linear_helper(xx, yy, zz, scale, weights):
    B = xx.shape[0]
    idx_range = torch.arange(0, B // 2, device=xx.device)
    s1 = idx_range * 2
    s2 = s1 + 1

    if scale.ndim == 0:
        scale = scale.unsqueeze(0)
    scale = scale.unsqueeze(1)
    weights = c_f.default(weights, get_default_kernel_weights(scale))

    loss1 = _mmd_linear(xx, s1, s2, scale, weights)
    loss2 = _mmd_linear(yy, s1, s2, scale, weights)
    loss3 = _mmd_linear(zz, s1, s2, scale, weights)
    loss4 = _mmd_linear(zz, s2, s1, scale, weights)

    return loss1, loss2, loss3, loss4


def get_mmd_linear(xx, yy, zz, scale, weights=None):
    # https://jmlr.csail.mit.edu/papers/volume13/gretton12a/gretton12a.pdf
    # https://arxiv.org/pdf/1502.02791.pdf
    is_joint_mmd = c_f.is_list_or_tuple(xx)
    B = xx[0].shape[0] if is_joint_mmd else xx.shape[0]

    if is_joint_mmd:
        product_list = [1, 1, 1, 1]
        for i in range(len(xx)):
            curr_kernels = _mmd_linear_helper(xx[i], yy[i], zz[i], scale[i], weights)
            product_list = [a * b for a, b in zip(product_list, curr_kernels)]
        loss1, loss2, loss3, loss4 = [torch.sum(a) for a in product_list]
    else:
        loss1, loss2, loss3, loss4 = [
            torch.sum(a) for a in _mmd_linear_helper(xx, yy, zz, scale, weights)
        ]

    loss = loss1 + loss2 - loss3 - loss4
    return torch.sum(loss) / float(B // 2)


def _mmd_quadratic_batched(rsum, scale, weights, query_is_ref):
    def fn(mat, s, *_):
        if query_is_ref:
            mat = c_f.mask_out_self(mat, s)
        rsum[0] += torch.sum(_mmd_quadratic(mat, scale, weights))

    return fn


def get_median_of_medians(x, dist_func):
    medians = []

    def fn(mat, *_):
        with torch.no_grad():
            medians.append(torch.median(mat))

    dist_func.iter_fn = fn
    dist_func(x, x)
    return torch.median(torch.stack(medians))


def get_mmd_quadratic_batched(x, y, dist_func, kernel_scales, bandwidth, weights=None):
    if torch.is_tensor(kernel_scales):
        kernel_scales = pml_cf.to_device(kernel_scales, x, dtype=x.dtype)
    if bandwidth is None:
        bandwidth = get_median_of_medians(x, dist_func)
    scale = -kernel_scales / bandwidth
    weights = c_f.default(weights, get_default_kernel_weights(scale))

    sums = []
    for s, t in [(x, x), (y, y), (x, y)]:
        rsum = [0]
        query_is_ref = s is t
        dist_func.iter_fn = _mmd_quadratic_batched(rsum, scale, weights, query_is_ref)
        dist_func(s, t)
        denom = (len(s) * (len(s) - 1)) if query_is_ref else (len(s) * len(t))
        sums.append(torch.sum(rsum[0]) / denom)

    return sums[0] + sums[1] - 2 * sums[2]



from typing import Any, List

import torch


class MultipleModels(torch.nn.Module):
    """
    Wraps a list of models, and returns their outputs as a list of tensors
    """

    def __init__(self, *models: torch.nn.Module):
        """
        Arguments:
            *models: The models to be wrapped.
        """
        super().__init__()
        self.models = torch.nn.ModuleList(models)

    def forward(self, x: torch.Tensor) -> List[Any]:
        """
        Arguments:
            x: the input to each model
        Returns:
            A list containing the output of each model.
        """
        outputs = []
        for m in self.models:
            outputs.append(m(x))
        return outputs



import math

import torch

from ..utils import common_functions as c_f


# adapted from https://github.com/jihanyang/AFN/blob/master/vanilla/Office31/SAFN/code/train.py
class AdaptiveFeatureNorm(torch.nn.Module):
    """
    Implementation of the loss in
    [Larger Norm More Transferable:
    An Adaptive Feature Norm Approach for
    Unsupervised Domain Adaptation](https://arxiv.org/abs/1811.07456).
    Encourages features to gradually have larger and larger L2 norms.
    """

    def __init__(self, step_size: float = 1):
        """
        Arguments:
            step_size: The desired increase in L2 norm at each iteration.
                Note that the loss will always be equal to ```step_size```
                because the goal is always to make the L2 norm ```step_size```
                larger than whatever the current L2 norm is.
        """
        super().__init__()
        self.step_size = step_size

    def forward(self, x):
        """"""
        l2_norm = x.norm(p=2, dim=1)
        radius = l2_norm.detach() + self.step_size
        return torch.mean((l2_norm - radius) ** 2)

    def extra_repr(self):
        """"""
        return c_f.extra_repr(self, ["step_size"])


class L2PreservedDropout(torch.nn.Module):
    """
    Implementation of the dropout layer described in
    [Larger Norm More Transferable:
    An Adaptive Feature Norm Approach for
    Unsupervised Domain Adaptation](https://arxiv.org/abs/1811.07456).
    Regular dropout preserves the L1 norm of features, whereas this
    layer preserves the L2 norm.
    """

    def __init__(self, p: float = 0.5, inplace: bool = False):
        """
        Arguments:
            p: probability of an element to be zeroed
            inplace: if set to True, will do this operation in-place
        """
        super().__init__()
        self.dropout = torch.nn.Dropout(p=p, inplace=inplace)
        self.scale = math.sqrt(1 - p)

    def forward(self, x):
        """"""
        x = self.dropout(x)
        if self.training:
            return x * self.scale
        return x



from typing import Callable

import torch

from ..utils import common_functions as c_f
from .entropy_loss import get_entropy
from .normalizers import SumNormalizer


def entropy_weights(logits, after_softmax, normalizer):
    entropies = get_entropy(logits, after_softmax)
    weights = 1 + torch.exp(-entropies)
    return normalizer(weights)


class EntropyWeights(torch.nn.Module):
    """
    Implementation of entropy weighting described in
    [Conditional Adversarial Domain Adaptation](https://arxiv.org/abs/1705.10667).
    Computes the entropy (```x```) per row of the input, and returns
    ```1+exp(-x)```.
    This can be used to weight losses, such that the most
    confidently scored samples have a higher weighting.
    """

    def __init__(
        self,
        after_softmax: bool = False,
        normalizer: Callable[[torch.Tensor], torch.Tensor] = None,
    ):
        """
        Arguments:
            after_softmax: If ```True```, then the rows of the input are assumed to
                already have softmax applied to them.
            normalizer: A callable for normalizing
                (e.g. min-max normalization) the weights.
                If ```None```, then sum normalization is used.
        """
        super().__init__()
        self.after_softmax = after_softmax
        self.normalizer = c_f.default(normalizer, SumNormalizer, {})

    def forward(self, logits: torch.Tensor) -> torch.Tensor:
        """
        Arguments:
            logits: Raw logits if ```self.after_softmax``` is False.
                Otherwise each row should be predictions that sum up to 1.
        """
        return entropy_weights(logits, self.after_softmax, self.normalizer)

    def extra_repr(self):
        """"""
        return c_f.extra_repr(self, ["after_softmax"])



from abc import abstractmethod

from .base_validator import BaseValidator


class SimpleLossValidator(BaseValidator):
    def __init__(self, layer="logits", **kwargs):
        super().__init__(**kwargs)
        self.layer = layer

    def compute_score(self, target_train):
        return -self.loss_fn(target_train[self.layer]).item()

    @property
    @abstractmethod
    def loss_fn(self, *args, **kwargs):
        pass



from typing import Callable, List

import torch

from ..layers import MMDLoss
from ..utils import common_functions as c_f
from .base import BaseHook, BaseWrapperHook
from .classification import CLossHook, SoftmaxLocallyHook
from .features import FeaturesAndLogitsHook, FeaturesHook
from .optimizer import OptimizerHook, SummaryHook
from .utils import ChainHook


class AlignerHook(BaseWrapperHook):
    """
    Computes an alignment loss (e.g MMD) based on features
    from two domains.
    """

    def __init__(
        self,
        loss_fn: Callable[[torch.Tensor, torch.Tensor], torch.Tensor] = None,
        hook: BaseHook = None,
        layer: str = "features",
        **kwargs,
    ):
        """
        Arguments:
            loss_fn: a function that computes a distance
                between two tensors. If ```None```,
                it defaults to [```MMDLoss```][pytorch_adapt.layers.MMDLoss].
            hook: the hook for computing features
            layer: the layer for which the loss is computed. Must be
                either ```"features"``` or ```"logits"```.
        """

        super().__init__(**kwargs)
        self.loss_fn = c_f.default(loss_fn, MMDLoss, {})
        if layer == "features":
            default_hook = FeaturesHook
        elif layer == "logits":
            default_hook = FeaturesAndLogitsHook
        else:
            raise ValueError("AlignerHook layer must be 'features' or 'logits'")
        self.hook = c_f.default(hook, default_hook, {})
        self.layer = layer

    def call(self, inputs, losses):
        outputs = self.hook(inputs, losses)[0]
        strs = c_f.filter(self.hook.out_keys, f"_{self.layer}$", ["^src", "^target"])
        [src, target] = c_f.extract([outputs, inputs], strs)
        confusion_loss = self.loss_fn(src, target)
        return outputs, {self._loss_keys()[0]: confusion_loss}

    def _loss_keys(self):
        return [f"{self.layer}_confusion_loss"]


class JointAlignerHook(BaseWrapperHook):
    """
    Computes a joint alignment loss (e.g Joint MMD) based on
    multiple features from two domains.

    The default setting is to use the features and logits
    from the source and target domains.
    """

    def __init__(
        self,
        loss_fn: Callable[
            [List[torch.Tensor], List[torch.Tensor]], torch.Tensor
        ] = None,
        hook: BaseHook = None,
        **kwargs,
    ):
        """
        Arguments:
            loss_fn: a function that computes a distance
                between two **lists** of tensors. If ```None```,
                it defaults to [```MMDLoss```][pytorch_adapt.layers.MMDLoss].
            hook: the hook for computing features and logits
        """
        super().__init__(**kwargs)
        self.loss_fn = c_f.default(loss_fn, MMDLoss, {})
        self.hook = c_f.default(hook, FeaturesAndLogitsHook, {})

    def call(self, inputs, losses):
        outputs = self.hook(inputs, losses)[0]
        src = self.get_all_domain_features(inputs, outputs, "src")
        target = self.get_all_domain_features(inputs, outputs, "target")
        confusion_loss = self.loss_fn(src, target)
        return outputs, {self._loss_keys()[0]: confusion_loss}

    def _loss_keys(self):
        return ["joint_confusion_loss"]

    def get_all_domain_features(self, inputs, outputs, domain):
        return c_f.extract(
            [outputs, inputs], sorted(c_f.filter(self.hook.out_keys, f"^{domain}"))
        )


class FeaturesLogitsAlignerHook(BaseWrapperHook):
    """
    This chains together an
    [```AlignerHook```][pytorch_adapt.hooks.AlignerHook] for
    ```"features"``` followed by an ```AlignerHook``` for ```"logits"```.
    """

    def __init__(
        self,
        loss_fn: Callable[[torch.Tensor, torch.Tensor], torch.Tensor] = None,
        **kwargs,
    ):
        """
        Arguments:
            loss_fn: The loss used by both aligner hooks.
        """
        super().__init__(**kwargs)
        loss_fn = c_f.default(loss_fn, MMDLoss, {})
        a1_hook = AlignerHook(loss_fn, layer="features")
        a2_hook = AlignerHook(loss_fn, layer="logits")
        self.hook = ChainHook(a1_hook, a2_hook)


class ManyAlignerHook(BaseWrapperHook):
    def __init__(
        self,
        loss_fn=None,
        aligner_hook=None,
        features_hook=None,
        softmax=True,
        **kwargs,
    ):
        super().__init__(**kwargs)
        loss_fn = c_f.default(loss_fn, MMDLoss, {})
        features_hook = c_f.default(features_hook, FeaturesAndLogitsHook, {})
        aligner_hook = c_f.default(
            aligner_hook, FeaturesLogitsAlignerHook, {"loss_fn": loss_fn}
        )
        if softmax:
            apply_to = c_f.filter(features_hook.out_keys, "_logits$")
            aligner_hook = SoftmaxLocallyHook(apply_to, aligner_hook)
        self.hook = ChainHook(features_hook, aligner_hook)


class AlignerPlusCHook(BaseWrapperHook):
    """
    Computes an alignment loss plus a classification loss,
    and then optimizes the models.
    """

    def __init__(
        self,
        opts,
        weighter=None,
        reducer=None,
        loss_fn=None,
        aligner_hook=None,
        pre=None,
        post=None,
        softmax=True,
        **kwargs,
    ):
        super().__init__(**kwargs)
        [pre, post] = c_f.many_default([pre, post], [[], []])
        aligner_hook = ManyAlignerHook(
            loss_fn=loss_fn, aligner_hook=aligner_hook, softmax=softmax
        )
        hook = ChainHook(*pre, aligner_hook, CLossHook(), *post)
        hook = OptimizerHook(hook, opts, weighter, reducer)
        s_hook = SummaryHook({"total_loss": hook})
        self.hook = ChainHook(hook, s_hook)



from ..layers import BNMLoss
from .simple_loss_validator import SimpleLossValidator


class BNMValidator(SimpleLossValidator):
    """
    Returns the negative of the
    [BNM loss][pytorch_adapt.layers.bnm_loss.BNMLoss]
    of all logits.
    """

    @property
    def loss_fn(self):
        return BNMLoss()



from typing import Any, Dict, List

from torch.utils.data import Dataset

from .domain_dataset import DomainDataset


class PseudoLabeledDataset(DomainDataset):
    """
    The wrapped dataset's ```__getitem__``` function
    should return a tuple of ```(data, label)```.
    The label will then be discarded,
    and the pseudo label will be returned instead.
    """

    def __init__(self, dataset: Dataset, pseudo_labels: List[int], domain: int = 0):
        """
        Arguments:
            dataset: The dataset to wrap
            pseudo_labels: The class labels that will be used
                instead of the labels contained in self.dataset
            domain: An integer representing the domain.
        """

        super().__init__(dataset, domain)
        if len(self.dataset) != len(pseudo_labels):
            raise ValueError("len(dataset) must equal len(pseudo_labels)")
        self.pseudo_labels = pseudo_labels

    def __len__(self):
        return len(self.dataset)

    def __getitem__(self, idx: int) -> Dict[str, Any]:
        """
        Returns:
            A dictionary with keys

                - "src_imgs" (the data)

                - "src_domain" (the integer representing the domain)

                - "src_labels" (the pseudo label)

                - "src_sample_idx" (idx)
        """

        img, _ = self.dataset[idx]
        return {
            "src_imgs": img,
            "src_domain": self.domain,
            "src_labels": self.pseudo_labels[idx],
            "src_sample_idx": idx,
        }



from typing import Tuple, Union

import ignite.distributed as idist
from ignite.engine import Engine, Events
from ignite.handlers import TerminateOnNan

from ...datasets import DataloaderCreator
from ...utils import common_functions as c_f
from ...validators import utils as val_utils
from .. import utils as f_utils
from . import checkpoint_utils
from . import utils as i_g


class Ignite:
    """
    Wraps an [Adapter][pytorch_adapt.adapters.BaseAdapter] and takes
    care of validation, model saving, etc. by using
    the event handler system of PyTorch Ignite.
    """

    def __init__(
        self,
        adapter,
        validator=None,
        val_hooks=None,
        checkpoint_fn=None,
        logger=None,
        log_freq=50,
        with_pbars=True,
        device=None,
        auto_dist=True,
        val_output_dict_fn=None,
    ):
        """
        Arguments:
            adapter: An [```Adapter```][pytorch_adapt.adapters.BaseAdapter] object, which contains
                the training and inference steps.
            validator: An optional [```ScoreHistory```][pytorch_adapt.validators.ScoreHistory] object,
                which is used to determine the best checkpoint.
            val_hooks: A list of functions that are called during validation. Each function should
                be one of the following types:

                - [```Validator```][pytorch_adapt.validators.BaseValidator]
                - [```ScoreHistory```][pytorch_adapt.validators.ScoreHistory]
                - A callable that accepts the following arguments:
                    - ```epoch``` (an integer)
                    - ```**kwargs```, keyword arguments mapping from dataset split name
                        to dictionaries containing features, labels etc.

                You can give your custom callable a list attribute named ```required_data```.
                For example, setting ```self.required_data = ["src_train", "target_train"]```,
                will ensure that the validation runner gathers data for the
                ```"src_train"``` and ```"target_train"``` splits.


            checkpoint_fn: A [```CheckpointFnCreator```][pytorch_adapt.frameworks.ignite.CheckpointFnCreator] object.
            logger: An object with the following functions:

                - ```add_training(adapter)```
                - ```add_validation(data, epoch)```, where ```data``` is ```{validator_name: validator}```,
                    and ```validator_name``` is usually just ```"validator"```.
                - ```write(engine)```

            log_freq: The number of iterations between logging
            with_pbars: If ```True```, progress bars are shown during each epoch.
            device: If ```None```, then it defaults to [```idist.device()```](https://pytorch.org/ignite/v0.4.8/distributed.html#ignite.distributed.utils.device).
            auto_dist: If ```True``` and ```device == None``` then
                [```auto_model```](https://pytorch.org/ignite/v0.4.8/generated/ignite.distributed.auto.auto_model.html)
                and [```auto_optim```](https://pytorch.org/ignite/v0.4.8/generated/ignite.distributed.auto.auto_optim.html) are applied
                to the models and optimizers.
        """
        self.adapter = adapter
        self.validator = validator
        self.val_hooks = c_f.default(val_hooks, [])
        self.checkpoint_fn = checkpoint_fn
        self.logger = logger
        self.log_freq = log_freq
        self.with_pbars = with_pbars
        self.device = c_f.default(device, idist.device, {})
        self.val_output_dict_fn = c_f.default(
            val_output_dict_fn, f_utils.create_output_dict
        )
        self.trainer_init()
        self.collector_init()
        self.dist_init_done = False
        if device is None and auto_dist:
            self.dist_init()
        self.temp_events = []

    def training_step(self, engine, batch):
        batch = c_f.batch_to_device(batch, self.device)
        return self.adapter.training_step(batch)

    def trainer_init(self):
        self.trainer = Engine(self.training_step)
        i_g.register(self.trainer, Events.STARTED, *self.trainer_started_events())
        i_g.register(
            self.trainer, Events.EPOCH_STARTED, *self.trainer_epoch_started_events()
        )
        i_g.register(
            self.trainer,
            Events.ITERATION_COMPLETED,
            *self.trainer_iteration_complete_events(),
        )
        i_g.register(
            self.trainer,
            Events.ITERATION_COMPLETED(every=self.log_freq),
            *self.trainer_log_freq_events(),
        )
        i_g.register(
            self.trainer, Events.EPOCH_COMPLETED, *self.trainer_epoch_complete_events()
        )

    def trainer_started_events(self):
        def fn(engine):
            self.adapter.before_training_starts(self)

        return [fn]

    def trainer_epoch_started_events(self):
        return [self.set_to_train(self.adapter.models)]

    def trainer_iteration_complete_events(self):
        output = [
            i_g.step_lr_schedulers(self.adapter.lr_schedulers, "per_step"),
            TerminateOnNan(),
        ]
        pbars = i_g.set_loggers_and_pbars(self, ["trainer"])
        if self.with_pbars:
            output.append(i_g.pbar_print_losses(pbars["trainer"]))
        return output

    def trainer_log_freq_events(self):
        output = []
        if self.logger:
            output.append(self.logger.add_training(self.adapter))
        return output

    def trainer_epoch_complete_events(self):
        output = [
            i_g.step_lr_schedulers(self.adapter.lr_schedulers, "per_epoch"),
            i_g.zero_grad(self.adapter),
        ]
        if self.logger:
            output.append(self.logger.write)
        return output

    def collector_init(self):
        self.collector = Engine(self.get_collector_step(self.adapter.inference))
        i_g.set_loggers_and_pbars(self, ["collector"])
        i_g.register(
            self.collector,
            Events.EPOCH_STARTED,
            self.set_to_eval(self.adapter.models),
        )

    def dist_init(self, *args, **kwargs):
        if not self.dist_init_done:
            self.adapter.models.apply(i_g.auto_model(*args, **kwargs))
            self.adapter.optimizers.apply(idist.auto_optim)
            self.dist_init_done = True

    def get_training_length(self):
        max_epochs = self.trainer.state.max_epochs
        max_iters = max_epochs * self.trainer.state.epoch_length
        return max_epochs, max_iters

    def get_all_outputs(self, dataloader, split_name):
        dataloaders = {split_name: dataloader}
        return i_g.collect_from_dataloaders(self.collector, dataloaders, [split_name])

    def run(
        self,
        datasets=None,
        dataloader_creator=None,
        dataloaders=None,
        val_interval=1,
        early_stopper_kwargs=None,
        resume=None,
        check_initial_score=False,
        **trainer_kwargs,
    ) -> Union[Tuple[float, int], Tuple[None, None]]:
        """
        Trains and optionally validates on the input datasets.

        Arguments:
            datasets: A dictionary mapping from dataset split names to datasets.
            dataloader_creator: Creates dataloaders using ```datasets```,
                when ```dataloaders == None```. Default value is
                [```DataloaderCreator()```][pytorch_adapt.datasets.DataloaderCreator].
            dataloaders: A dictionary mapping from dataset split names to
                dataloaders. If ```None```, then ```datasets``` must be provided.
            val_interval: ```self.validator``` and ```self.val_hooks``` will be called
                every ```val_interval``` epochs.
            early_stopper_kwargs: A dictionary of keyword arguments to be passed to
                [```EarlyStopping```](https://pytorch.org/ignite/v0.4.8/generated/ignite.handlers.early_stopping.EarlyStopping.html).
            resume: Resume training from a checkpoint. This should be either
                a string representing a file path, or an integer representing a global step.
            check_initial_score: If ```True```, then ```self.validator``` and ```self.val_hooks```
                will be called before training starts. For example, you might use this to check the performance of
                a pretrained model before finetuning.
            **trainer_kwargs: Keyword arguments that are passed to the PyTorch Ignite
                [```run()```](https://pytorch.org/ignite/v0.4.8/generated/ignite.engine.engine.Engine.html#ignite.engine.engine.Engine.run)
                function.
        Returns:
            A tuple of ```(best_score, best_epoch)``` or ```(None, None)```
                if no validator is used.
        """
        if dataloaders is None:
            dataloader_creator = c_f.default(dataloader_creator, DataloaderCreator())
            dataloaders = dataloader_creator(**datasets)

        self.remove_temp_events()
        max_epochs = trainer_kwargs.get("max_epochs", 1)
        condition = i_g.interval_condition(val_interval, max_epochs)

        if self.checkpoint_fn:
            self.add_checkpoint_fn(condition, dataloaders)
        elif self.validator or self.val_hooks:
            self.add_validation_runner(condition, dataloaders)

        if check_initial_score:
            self.add_validation_runner(Events.STARTED, dataloaders)

        if self.validator and early_stopper_kwargs:
            self.add_early_stopper(val_interval, **early_stopper_kwargs)

        if resume is not None:
            self.load_checkpoint(resume)

        if not i_g.is_done(self.trainer, max_epochs):
            self.trainer.run(dataloaders["train"], **trainer_kwargs)

        self.remove_temp_events()

        if self.validator:
            return self.validator.best_score, self.validator.best_epoch

        return None, None

    def get_validation_runner(
        self,
        dataloaders,
    ):
        return i_g.get_validation_runner(
            self.collector,
            dataloaders,
            self.validator,
            self.val_hooks,
            self.logger,
        )

    def add_validation_runner(self, condition, dataloaders):
        val_runner = self.get_validation_runner(dataloaders)
        self.add_temp_event_handler(condition, val_runner)

    def add_checkpoint_fn(self, condition, dataloaders):
        score_function = (
            self.get_validation_runner(dataloaders) if self.validator else None
        )
        self.add_temp_event_handler(
            condition,
            self.checkpoint_fn(
                adapter=self.adapter,
                validator=self.validator,
                val_hooks=self.val_hooks,
                score_function=score_function,
            ),
        )
        if not self.validator and self.val_hooks:
            self.add_validation_runner(condition, dataloaders)

    def add_early_stopper(self, val_interval, **kwargs):
        def score_fn(_):
            return self.validator.latest_score

        self.add_temp_event_handler(
            Events.EPOCH_COMPLETED(every=val_interval),
            i_g.early_stopper(**kwargs)(
                trainer=self.trainer,
                score_function=score_fn,
            ),
        )

    def load_checkpoint(self, resume):
        to_load = {
            "engine": self.trainer,
            "validator": self.validator,
            **checkpoint_utils.adapter_to_dict(self.adapter),
            **checkpoint_utils.val_hooks_to_dict(self.val_hooks),
        }
        if isinstance(resume, str):
            kwargs = {"checkpoint": resume}
        elif isinstance(resume, int):
            kwargs = {"global_step": resume}
        else:
            raise TypeError(
                "resume must be a string representing a file path, or an integer representing a global step"
            )
        self.checkpoint_fn.load_objects(to_load, **kwargs)
        i_g.resume_checks(self.trainer, self.validator)

    def evaluate_best_model(
        self, datasets, validator, dataloader_creator=None
    ) -> float:
        """
        Loads the best checkpoint and computes the score on the given datasets.
            This requires ```self.checkpoint_fn``` to be not ```None```.
        Arguments:
            datasets: A dictionary mapping from dataset split names to datasets.
            validator: A [Validator][pytorch_adapt.validators.BaseValidator]
                or [ScoreHistory][pytorch_adapt.validators.ScoreHistory] object,
                which will compute and return the score.
            dataloader_creator: If ```None```, it will default to
                [```DataloaderCreator()```][pytorch_adapt.datasets.DataloaderCreator].
        Returns:
            The validator's score for the best checkpoint.
        """
        c_f.LOGGER.info("***EVALUATING BEST MODEL***")
        dataloader_creator = c_f.default(dataloader_creator, DataloaderCreator, {})
        dataloaders = dataloader_creator(**datasets)
        self.checkpoint_fn.load_best_checkpoint({"models": self.adapter.models})
        collected_data = i_g.collect_from_dataloaders(
            self.collector, dataloaders, validator.required_data
        )
        return val_utils.call_val_hook(validator, collected_data)

    def get_collector_step(self, inference):
        def collector_step(engine, batch):
            batch = c_f.batch_to_device(batch, self.device)
            return f_utils.collector_step(inference, batch, self.val_output_dict_fn)

        return collector_step

    def set_to_train(self, models):
        def handler(engine):
            c_f.LOGGER.info("Setting models to train() mode")
            models.train()

        return handler

    def set_to_eval(self, models):
        def handler(engine):
            c_f.LOGGER.info("Setting models to eval() mode")
            models.eval()

        return handler

    def add_temp_event_handler(self, event, handler):
        removable = self.trainer.add_event_handler(event, handler)
        self.temp_events.append(removable)

    def remove_temp_events(self):
        for h in self.temp_events:
            h.remove()
        self.temp_events = []



from .inference import (
    adabn_fn,
    adda_fn,
    adda_full_fn,
    adda_with_d,
    cdan_full_fn,
    d_bridge_fn,
    default_fn,
    default_with_d,
    default_with_d_logits_layer,
    gvb_full_fn,
    gvb_with_g_bridge,
    mcd_fn,
    mcd_full_fn,
    rtn_fn,
    rtn_full_fn,
    rtn_with_feature_combiner,
    symnets_fn,
    symnets_full_fn,
    with_d,
    with_d_bridge,
    with_feature_combiner,
)



from typing import Union

from ..utils import common_functions as c_f
from .base_container import BaseContainer
from .lr_schedulers import LRSchedulers
from .misc import Misc
from .models import Models
from .optimizers import Optimizers


def get_container(k):
    if k == "models":
        return Models({})
    elif k == "optimizers":
        return Optimizers({})
    elif k == "lr_schedulers":
        return LRSchedulers({})
    elif k == "misc":
        return Misc({})
    else:
        return BaseContainer({})


class MultipleContainers(BaseContainer):
    """
    Contains other containers and initializes them.
    """

    def __init__(self, **kwargs):
        self.store = kwargs

    def merge(self, **kwargs: Union[BaseContainer, None]):
        """
        Merges the input containers into any existing sub-containers.
        """
        for k, v in kwargs.items():
            if isinstance(v, BaseContainer):
                if k in self:
                    self[k].merge(v)
                else:
                    self[k] = v
            elif v is None:
                if k not in self:
                    self[k] = get_container(k)
            else:
                raise TypeError(
                    f"Input to {c_f.cls_name(self)}.merge must be BaseContainer or None"
                )

    def create(self):
        """
        Calls [```.create()```][pytorch_adapt.containers.BaseContainer.create]
        or [```.create_with()```][pytorch_adapt.containers.BaseContainer.create_with]
        on sub-containers.

        - Optimizers are created with models as input.
        - LR schedulers are created with optimizers as input.
        """
        self["models"].create()
        self["optimizers"].create_with(self["models"])
        self["lr_schedulers"].create_with(self["optimizers"])
        self["misc"].create()



from typing import Any, Dict

from torch.utils.data import Dataset

from .domain_dataset import DomainDataset


class TargetDataset(DomainDataset):
    """
    Wrap your target dataset with this.

    If ```supervised = True```, the wrapped dataset's ```__getitem__```
    must return a tuple of ```(data, label)```.
    Otherwise it can return either ```(data, label)``` or ```data```.
    """

    def __init__(self, dataset: Dataset, domain: int = 1, supervised: bool = False):
        """
        Arguments:
            dataset: The dataset to wrap
            domain: An integer representing the domain.
            supervised: A boolean for if the target dataset should return labels.
        """
        super().__init__(dataset, domain)
        self.supervised = supervised

    def __getitem__(self, idx: int) -> Dict[str, Any]:
        """
        Returns:
            A dictionary with keys

                - "target_imgs" (the data)

                - "target_domain" (the integer representing the domain)

                - "target_sample_idx" (idx)

                If ```supervised = True``` it returns an extra key

                - "target_labels" (the class label)
        """

        has_labels = False
        img = self.dataset[idx]
        if isinstance(img, (list, tuple)):
            has_labels = True
            img, labels = img

        if self.supervised and not has_labels:
            raise ValueError(
                "if TargetDataset is instantiated with supervised=True, the wrapped dataset must include labels"
            )

        item = {
            "target_imgs": img,
            "target_domain": self.domain,
            "target_sample_idx": idx,
        }

        if self.supervised:
            item["target_labels"] = labels

        return item



from ...validators import utils as val_utils


class IgniteValHookWrapper:
    def __init__(self, validator, logger=None):
        self.validator = validator
        self.logger = logger

    def __call__(self, epoch, **kwargs):
        val_utils.call_val_hook(self.validator, kwargs, epoch)
        if self.logger:
            self.logger.add_validation({"val_hook": self.validator}, epoch)

    @property
    def required_data(self):
        return self.validator.required_data

    def state_dict(self):
        return self.validator.state_dict()

    def load_state_dict(self, state_dict):
        self.validator.load_state_dict(state_dict)



from .classifier import Classifier
from .discriminator import Discriminator
from .mnist import MNISTFeatures
from .pretrained import (
    domainnet126C,
    domainnet126G,
    mnistC,
    mnistG,
    office31C,
    office31G,
    officehomeC,
    officehomeG,
    voc_multilabelC,
    voc_multilabelG,
)
from .pretrained_scores import pretrained_src_accuracy, pretrained_target_accuracy



from torchvision.datasets import MNIST

from ..transforms.classification import get_mnist_transform, get_resnet_transform
from ..utils import common_functions as c_f
from .clipart1k import Clipart1kMultiLabel
from .combined_source_and_target import CombinedSourceAndTargetDataset
from .concat_dataset import ConcatDataset
from .domainnet import DomainNet126
from .mnistm import MNISTM
from .office31 import Office31
from .officehome import OfficeHome
from .source_dataset import SourceDataset
from .target_dataset import TargetDataset
from .voc_multilabel import VOCMultiLabel
from .voc_multilabel import get_labels_as_vector as voc_labels_as_vector


def get_multiple(dataset_getter, domains, **kwargs):
    return ConcatDataset([dataset_getter(domain=d, **kwargs) for d in domains])


def get_datasets(
    dataset_getter,
    src_domains,
    target_domains,
    folder,
    download=False,
    return_target_with_labels=False,
    supervised=False,
    transform_getter=None,
    **kwargs,
):
    def getter(domains, train, is_training):
        return get_multiple(
            dataset_getter,
            domains,
            train=train,
            is_training=is_training,
            root=folder,
            download=download,
            transform_getter=transform_getter,
            **kwargs,
        )

    if not src_domains and not target_domains:
        raise ValueError(
            "At least one of src_domains and target_domains must be provided"
        )

    output = {}
    if src_domains:
        output["src_train"] = SourceDataset(getter(src_domains, True, False))
        output["src_val"] = SourceDataset(getter(src_domains, False, False))
    if target_domains:
        output["target_train"] = TargetDataset(
            getter(target_domains, True, False), supervised=supervised
        )
        output["target_val"] = TargetDataset(
            getter(target_domains, False, False), supervised=supervised
        )
        # For academic setting: unsupervised learning w/ seperate target datasets that have gt lables for eval.
        if return_target_with_labels:
            output["target_train_with_labels"] = TargetDataset(
                getter(target_domains, True, False), domain=1, supervised=True
            )
            output["target_val_with_labels"] = TargetDataset(
                getter(target_domains, False, False), domain=1, supervised=True
            )
    if src_domains and target_domains:
        output["train"] = CombinedSourceAndTargetDataset(
            SourceDataset(getter(src_domains, True, True)),
            TargetDataset(getter(target_domains, True, True)),
        )
    elif src_domains:
        output["train"] = SourceDataset(getter(src_domains, True, True))
    elif target_domains:
        output["train"] = TargetDataset(getter(target_domains, True, True))
    return output


def _get_mnist_mnistm(is_training, transform_getter, **kwargs):
    transform_getter = c_f.default(transform_getter, get_mnist_transform)
    domain = kwargs["domain"]
    kwargs["transform"] = transform_getter(
        domain=domain, train=kwargs["train"], is_training=is_training
    )
    kwargs.pop("domain")
    if domain == "mnist":
        return MNIST(**kwargs)
    elif domain == "mnistm":
        return MNISTM(**kwargs)


def get_mnist_mnistm(*args, **kwargs):
    return get_datasets(_get_mnist_mnistm, *args, **kwargs)


def standard_dataset(cls):
    def fn(is_training, transform_getter, **kwargs):
        transform_getter = c_f.default(transform_getter, get_resnet_transform)
        kwargs["transform"] = transform_getter(
            domain=kwargs["domain"], train=kwargs["train"], is_training=is_training
        )
        return cls(**kwargs)

    return fn


def get_office31(*args, **kwargs):
    return get_datasets(standard_dataset(Office31), *args, **kwargs)


def get_officehome(*args, **kwargs):
    return get_datasets(standard_dataset(OfficeHome), *args, **kwargs)


def get_domainnet126(*args, **kwargs):
    return get_datasets(standard_dataset(DomainNet126), *args, **kwargs)


def _get_voc_multilabel(is_training, transform_getter, **kwargs):
    # import here, because albumentations is an optional dependency
    from ..transforms.detection import VOCTransformWrapper, get_voc_transform

    transform_getter = c_f.default(transform_getter, get_voc_transform)
    domain = kwargs["domain"]
    transform = transform_getter(
        domain=domain, train=kwargs["train"], is_training=is_training
    )
    kwargs["transforms"] = VOCTransformWrapper(transform, voc_labels_as_vector)
    kwargs.pop("domain")
    train = kwargs.pop("train")
    if domain == "voc":
        kwargs["image_set"] = "train" if train else "val"
        return VOCMultiLabel(**kwargs)
    elif domain == "clipart":
        kwargs.pop("year", None)
        kwargs["image_set"] = "train" if train else "test"
        return Clipart1kMultiLabel(**kwargs)


def get_voc_multilabel(*args, **kwargs):
    return get_datasets(_get_voc_multilabel, *args, **kwargs)



import logging
import os
import shutil

import numpy as np
import torch
from pytorch_metric_learning.utils import common_functions as pml_cf
from sklearn.model_selection import train_test_split
from torchmetrics.functional import accuracy as tmf_accuracy

from ..adapters import Finetuner
from ..containers import Models, Optimizers
from ..datasets import DataloaderCreator, SourceDataset
from ..models import Discriminator
from ..utils import common_functions as c_f
from .accuracy_validator import AccuracyValidator
from .base_validator import BaseValidator
from .score_history import ScoreHistory


def dev_binary_fn(preds, labels):
    preds = torch.argmax(preds, dim=1)
    return (preds != labels).float()


def default_framework_fn(adapter, validator, folder):
    from ..frameworks.ignite import CheckpointFnCreator, Ignite

    validator = ScoreHistory(validator)
    checkpoint_fn = CheckpointFnCreator(dirname=folder)
    return Ignite(
        adapter, validator=validator, checkpoint_fn=checkpoint_fn, with_pbars=False
    )


class DeepEmbeddedValidator(BaseValidator):
    """
    Implementation of
    [Towards Accurate Model Selection in Deep Unsupervised Domain Adaptation](http://proceedings.mlr.press/v97/you19a.html)
    """

    def __init__(
        self,
        temp_folder,
        layer="features",
        num_workers=0,
        batch_size=32,
        error_fn=None,
        error_layer="logits",
        normalization=None,
        framework_fn=None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.temp_folder = temp_folder
        self.layer = layer
        self.num_workers = num_workers
        self.batch_size = batch_size
        self.error_fn = c_f.default(
            error_fn, torch.nn.CrossEntropyLoss(reduction="none")
        )
        self.error_layer = error_layer
        check_normalization(normalization)
        self.normalization = normalization
        self.framework_fn = c_f.default(framework_fn, default_framework_fn)
        self.D_accuracy_val = None
        self.D_accuracy_test = None
        self.mean_error = None
        self._DEV_recordable = ["D_accuracy_val", "D_accuracy_test", "mean_error"]
        pml_cf.add_to_recordable_attributes(self, list_of_names=self._DEV_recordable)

    def compute_score(self, src_train, src_val, target_train):
        init_logging_level = c_f.LOGGER.level
        c_f.LOGGER.setLevel(logging.WARNING)
        weights, self.D_accuracy_val, self.D_accuracy_test = get_weights(
            src_train[self.layer],
            src_val[self.layer],
            target_train[self.layer],
            self.num_workers,
            self.batch_size,
            self.temp_folder,
            self.framework_fn,
        )
        error_per_sample = self.error_fn(src_val[self.error_layer], src_val["labels"])
        output = get_dev_risk(weights, error_per_sample[:, None], self.normalization)
        self.mean_error = torch.mean(error_per_sample).item()
        c_f.LOGGER.setLevel(init_logging_level)
        return -output

    def extra_repr(self):
        x = super().extra_repr()
        x += f"\n{c_f.extra_repr(self, self._DEV_recordable)}"
        return x


#########################################################################
#### ADAPTED FROM https://github.com/thuml/Deep-Embedded-Validation #####
#########################################################################


def check_normalization(normalization):
    if normalization not in [None, "max", "standardize"]:
        raise ValueError("normalization must be one of [None, 'max', 'standardize']")


def normalize_weights(weights, normalization):
    check_normalization(normalization)
    if normalization == "max":
        weights /= np.max(weights)  # normalize between 0 and 1
        weights -= np.mean(weights) - 1  # shift to have mean of 1
    elif normalization == "standardize":
        weights = (weights - np.mean(weights)) / np.std(weights)  # standardize
        weights += 1  # shift to have mean of 1
    return weights


def get_dev_risk(weight, error, normalization):
    weight = weight.cpu().numpy()
    error = error.cpu().numpy()

    N, d = weight.shape
    _N, _d = error.shape
    assert N == _N and d == _d, "dimension mismatch!"

    weight = normalize_weights(weight, normalization)
    weighted_error = weight * error
    cov = np.cov(np.concatenate((weighted_error, weight), axis=1), rowvar=False)[0][1]
    var_w = np.var(weight, ddof=1)
    eta = -cov / var_w
    return np.mean(weighted_error) + eta * np.mean(weight) - eta


def get_weights(
    source_feature,
    validation_feature,
    target_feature,
    num_workers,
    batch_size,
    temp_folder,
    framework_fn,
):
    device = source_feature.device
    source_feature = source_feature.cpu().numpy()
    validation_feature = validation_feature.cpu().numpy()
    target_feature = target_feature.cpu().numpy()

    N_s, d = source_feature.shape
    N_t, _d = target_feature.shape
    source_feature = source_feature.copy()
    target_feature = target_feature.copy()
    all_feature = np.concatenate((source_feature, target_feature))
    all_label = np.asarray([1] * N_s + [0] * N_t, dtype=np.int64)
    (
        feature_for_train,
        feature_for_test,
        label_for_train,
        label_for_test,
    ) = train_test_split(all_feature, all_label, train_size=0.8)

    train_set = SourceDataset(
        pml_cf.EmbeddingDataset(feature_for_train, label_for_train)
    )
    val_set = SourceDataset(pml_cf.EmbeddingDataset(feature_for_test, label_for_test))

    decays = [1e-1, 3e-2, 1e-2, 3e-3, 1e-3, 3e-4, 1e-4, 3e-5, 1e-5]
    val_acc, trainers, folders = [], [], []
    epochs = 100
    patience = 2

    for i, decay in enumerate(decays):
        torch.cuda.empty_cache()
        curr_folder = os.path.join(temp_folder, f"DeepEmbeddedValidation{i}")
        models = Models(
            {
                "G": torch.nn.Identity(),
                "C": Discriminator(d, h=d, out_size=2).to(device),
            }
        )
        optimizers = Optimizers(
            (torch.optim.Adam, {"lr": 0.001, "weight_decay": decay})
        )
        trainer = Finetuner(models=models, optimizers=optimizers)
        validator = AccuracyValidator(
            torchmetric_kwargs={"average": "macro", "num_classes": 2}
        )
        trainer = framework_fn(trainer, validator, curr_folder)
        datasets = {"train": train_set, "src_val": val_set}
        bs = int(np.min([len(train_set), len(val_set), batch_size]))

        acc, _ = trainer.run(
            datasets,
            dataloader_creator=DataloaderCreator(
                num_workers=num_workers, batch_size=bs
            ),
            max_epochs=epochs,
            val_interval=1,
            early_stopper_kwargs={"patience": patience},
        )
        val_acc.append(acc)
        trainers.append(trainer)
        folders.append(curr_folder)

    torch.cuda.empty_cache()
    D_accuracy_val = max(val_acc)
    index = val_acc.index(D_accuracy_val)

    labels = torch.ones(len(validation_feature), dtype=int)
    validation_set = SourceDataset(pml_cf.EmbeddingDataset(validation_feature, labels))
    trainer = trainers[index]
    trainer.checkpoint_fn.load_best_checkpoint(
        {"models": trainer.adapter.models},
    )
    bs = min(len(validation_set), batch_size)
    dataloader = torch.utils.data.DataLoader(
        validation_set, num_workers=num_workers, batch_size=bs
    )
    domain_out = trainer.get_all_outputs(dataloader, "val")
    domain_out = domain_out["val"]["preds"]
    weights = (domain_out[:, :1] / domain_out[:, 1:]) * (float(N_s) / N_t)

    [shutil.rmtree(f) for f in folders]

    D_accuracy_test = tmf_accuracy(domain_out, labels.to(domain_out.device)).item()

    return weights, D_accuracy_val, D_accuracy_test



import torch

from ..containers import KeyEnforcer
from ..hooks import (
    CDANEHook,
    CDANHook,
    DomainConfusionHook,
    GANEHook,
    GANHook,
    VADAHook,
)
from .base_adapter import BaseGCDAdapter
from .utils import with_opt


class GAN(BaseGCDAdapter):
    """
    Wraps [GANHook][pytorch_adapt.hooks.GANHook].

    |Container|Required keys|
    |---|---|
    |models|```["G", "C", "D"]```|
    |optimizers|```["G", "C", "D"]```|
    """

    def init_hook(self, hook_kwargs):
        g_opts = with_opt(["G", "C"])
        d_opts = with_opt(["D"])
        self.hook = self.hook_cls(d_opts=d_opts, g_opts=g_opts, **hook_kwargs)

    @property
    def hook_cls(self):
        return GANHook


class GANE(GAN):
    @property
    def hook_cls(self):
        return GANEHook


class CDAN(GAN):
    """
    Wraps [CDANHook][pytorch_adapt.hooks.CDANHook].

    |Container|Required keys|
    |---|---|
    |models|```["G", "C", "D"]```|
    |optimizers|```["G", "C", "D"]```|
    |misc|```["feature_combiner"]```|
    """

    @property
    def hook_cls(self):
        return CDANHook

    def get_key_enforcer(self) -> KeyEnforcer:
        ke = super().get_key_enforcer()
        ke.requirements["misc"] = ["feature_combiner"]
        return ke


class CDANE(CDAN):
    @property
    def hook_cls(self):
        return CDANEHook


class DomainConfusion(GAN):
    """
    Wraps [DomainConfusionHook][pytorch_adapt.hooks.DomainConfusionHook].

    |Container|Required keys|
    |---|---|
    |models|```["G", "C", "D"]```|
    |optimizers|```["G", "C", "D"]```|
    """

    @property
    def hook_cls(self):
        return DomainConfusionHook


class VADA(GAN):
    """
    Wraps [VADAHook][pytorch_adapt.hooks.VADAHook].

    |Container|Required keys|
    |---|---|
    |models|```["G", "C", "D"]```|
    |optimizers|```["G", "C", "D"]```|
    |misc|```["combined_model"]```|

    The ```"combined_model"``` key does not need to be passed in.
    It is simply ```torch.nn.Sequential(G, C)```, and is set
    automatically during initialization.
    """

    @property
    def hook_cls(self):
        return VADAHook

    def init_containers_and_check_keys(self, containers):
        models = containers["models"]
        misc = containers["misc"]
        misc["combined_model"] = torch.nn.Sequential(models["G"], models["C"])
        super().init_containers_and_check_keys(containers)

    def get_key_enforcer(self) -> KeyEnforcer:
        ke = super().get_key_enforcer()
        ke.requirements["misc"] = ["combined_model"]
        return ke



import torch

from ..utils import common_functions as c_f


def create_output_dict(f_dict):
    return {**f_dict, "preds": torch.softmax(f_dict["logits"], dim=1)}


def create_output_dict_multilabel_classification(f_dict):
    return {**f_dict, "preds": torch.sigmoid(f_dict["logits"])}


def create_output_dict_preds_as_features(f_dict):
    [features, logits] = c_f.extract(f_dict, ["features", "logits"])
    if not torch.allclose(features, logits):
        raise ValueError("features and logits should be equal")
    return {"features": features, "preds": features}


def extract_data(batch):
    data = {"_".join(k.split("_")[1:]): v for k, v in batch.items()}
    if len(data) != len(batch):
        raise KeyError(
            f"Batch should have only one domain, but it has keys: {batch.keys()}"
        )
    return data


def collector_step(inference, batch, output_dict_fn):
    data = extract_data(batch)
    with torch.no_grad():
        f_dict = inference(data["imgs"], domain=data["domain"])
    data.pop("imgs")  # we don't want to collect imgs
    f_dict = output_dict_fn(f_dict)
    if len(f_dict.keys() & data.keys()) > 0:
        raise ValueError("f_dict and data should have no overlap at this point")
    f_dict.update(data)
    return f_dict


def filter_datasets(datasets, validator):
    return {
        k: v for k, v in datasets.items() if k in ["train"] + validator.required_data
    }



from typing import Callable

import torch

from ..utils import common_functions as c_f
from .entropy_weights import EntropyWeights
from .normalizers import SumNormalizer


# reference https://github.com/thuml/Versatile-Domain-Adaptation
class MCCLoss(torch.nn.Module):
    """
    Implementation of
    [Minimum Class Confusion for Versatile Domain Adaptation](https://arxiv.org/abs/1912.03699).
    """

    def __init__(
        self,
        T: float = 1,
        entropy_weighter: Callable[[torch.Tensor], torch.Tensor] = None,
    ):
        """
        Arguments:
            T: softmax temperature applied to the input target logits
            entropy_weighter: a function that returns a weight for each
                sample. The weights are used in the process of computing
                the class confusion tensor as described in the paper.
                If ```None```, then ```layers.EntropyWeights``` is used.
        """
        super().__init__()
        self.T = T
        self.entropy_weighter = c_f.default(
            entropy_weighter,
            EntropyWeights(
                after_softmax=True, normalizer=SumNormalizer(scale_by_batch_size=True)
            ),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Arguments:
            x: target logits
        """
        Y = torch.nn.functional.softmax(x / self.T, dim=1)
        H_weights = self.entropy_weighter(Y.detach())
        C = torch.mm((Y * H_weights.view(-1, 1)).t(), Y)
        C = C / torch.sum(C, dim=1)
        return (torch.sum(C) - torch.trace(C)) / C.shape[0]

    def extra_repr(self):
        """"""
        return c_f.extra_repr(self, ["T"])



from .base_dataset import BaseDataset, BaseDownloadableDataset
from .clipart1k import Clipart1kMultiLabel
from .combined_source_and_target import CombinedSourceAndTargetDataset
from .concat_dataset import ConcatDataset
from .dataloader_creator import DataloaderCreator
from .domainnet import DomainNet, DomainNet126, DomainNet126Full
from .getters import (
    get_domainnet126,
    get_mnist_mnistm,
    get_office31,
    get_officehome,
    get_voc_multilabel,
)
from .mnistm import MNISTM
from .office31 import Office31, Office31Full
from .officehome import OfficeHome, OfficeHomeFull
from .pseudo_labeled_dataset import PseudoLabeledDataset
from .source_dataset import SourceDataset
from .target_dataset import TargetDataset
from .voc_multilabel import VOCMultiLabel



import inspect
from abc import ABC, abstractmethod
from typing import Dict, List

from ..utils import common_functions as c_f


class BaseValidator(ABC):
    """
    The parent class of all validators.

    The main purpose of validators is to give an estimate
    of target domain accuracy, usually without having access to
    class labels.
    """

    def __init__(self, key_map: Dict[str, str] = None):
        """
        Arguments:
            key_map: A mapping from ```<new_split_names>``` to
                ```<original_split_names>```. For example,
                [```AccuracyValidator```][pytorch_adapt.validators.AccuracyValidator]
                expects ```src_val``` by default. When used with one of the
                [```frameworks```](../frameworks/index.md), this default
                indicates that data related to the ```src_val``` split should be retrieved.
                If you instead want to compute accuracy for the ```src_train``` split,
                you would set the ```key_map``` to ```{"src_train": "src_val"}```.
        """
        self.key_map = c_f.default(key_map, {})

    def _required_data(self):
        args = inspect.getfullargspec(self.compute_score).args
        args.remove("self")
        return args

    @property
    def required_data(self) -> List[str]:
        """
        Returns:
            A list of dataset split names.
        """
        output = set(self._required_data()) - set(self.key_map.values())
        output = list(output)
        for k, v in self.key_map.items():
            output.append(k)
        return output

    @abstractmethod
    def compute_score(self):
        pass

    def __call__(self, **kwargs) -> float:
        """
        Arguments:
            **kwargs: A mapping from dataset split name to
                dictionaries containing:

                - ```"features"```
                - ```"logits"```
                - ```"preds"```
                - ```"domain"```
                - ```"labels"``` (if available)
        Returns:
            The validation score.
        """
        kwargs = self.kwargs_check(kwargs)
        return self.compute_score(**kwargs)

    def kwargs_check(self, kwargs):
        if kwargs.keys() != set(self.required_data):
            raise ValueError(
                f"Input to compute_score has keys = {kwargs.keys()} but should have keys {self.required_data}"
            )
        return c_f.map_keys(kwargs, self.key_map)

    def __repr__(self):
        return c_f.nice_repr(self, self.extra_repr(), {})

    def extra_repr(self):
        return c_f.extra_repr(self, ["required_data"])



from .base_container import BaseContainer, DeleteKey
from .key_enforcer import KeyEnforcer
from .lr_schedulers import LRSchedulers
from .misc import Misc
from .models import Models
from .multiple_containers import MultipleContainers
from .optimizers import Optimizers



import torch
from pytorch_metric_learning.utils import common_functions as pml_cf

from ....hooks.base import BaseHook
from ....utils import common_functions as c_f
from ....weighters.base_weighter import BaseWeighter


class IgniteRecordKeeperLogger:
    """
    Uses [record-keeper](https://github.com/KevinMusgrave/record-keeper)
    to record data tensorboard, csv, and sqlite.
    """

    def __init__(
        self,
        folder=None,
        tensorboard_writer=None,
        record_writer=None,
        attr_list_names=None,
    ):
        """
        Arguments:
            folder: path where records will be saved.
            tensorboard_writer:
            record_writer: a ```RecordWriter``` object (see record-keeper)
        """
        from record_keeper import RecordKeeper, RecordWriter
        from torch.utils.tensorboard import SummaryWriter

        tensorboard_writer = c_f.default(
            tensorboard_writer,
            SummaryWriter,
            {"log_dir": folder, "max_queue": 1000000, "flush_secs": 30},
        )
        record_writer = c_f.default(record_writer, RecordWriter, {"folder": folder})
        attr_list_names = c_f.default(
            attr_list_names, pml_cf.list_of_recordable_attributes_list_names()
        )
        self.record_keeper = RecordKeeper(
            tensorboard_writer=tensorboard_writer,
            record_writer=record_writer,
            attributes_to_search_for=attr_list_names,
        )

    def add_training(self, adapter):
        def fn(engine):
            record_these = [
                ({"engine_output": engine.state.output}, {}),
                (
                    adapter.optimizers,
                    {
                        "parent_name": "optimizers",
                        "custom_attr_func": optimizer_attr_func,
                    },
                ),
                ({"misc": adapter.misc}, {}),
                (
                    {"hook": adapter.hook},
                    {"recursive_types": [BaseHook, BaseWeighter, torch.nn.Module]},
                ),
            ]
            for record, kwargs in record_these:
                self.record_keeper.update_records(
                    record, engine.state.iteration, **kwargs
                )

        return fn

    def add_validation(self, data, epoch):
        self.record_keeper.update_records(data, epoch)

    def write(self, engine):
        self.record_keeper.save_records()
        self.record_keeper.tensorboard_writer.flush()


def optimizer_attr_func(optimizer):
    return {"lr": c_f.get_lr(optimizer)}



import glob
import os

from ignite.handlers import ModelCheckpoint

from ...adapters.utils import container_names
from ...utils import common_functions as c_f


def global_step_transform(engine, _):
    return engine.state.epoch


def val_hooks_to_dict(val_hooks):
    output = {}
    for i, v in enumerate(val_hooks):
        if not all(hasattr(v, x) for x in ["state_dict", "load_state_dict"]):
            c_f.LOGGER.warning(
                "val_hook has no state_dict or load_state_dict so it will not be saved or loaded"
            )
        else:
            output[f"val_hook{i}"] = v
    return output


def adapter_to_dict(adapter):
    return {k: getattr(adapter, k) for k in container_names()}


class CheckpointFnCreator:
    """
    This class creates a checkpointing function
    for use with the [```Ignite```][pytorch_adapt.frameworks.ignite.Ignite] wrapper.
    """

    def __init__(self, **kwargs):
        """
        Arguments:
            **kwargs: Optional arguments that will be passed to PyTorch Ignite's
                [```ModelCheckpoint```](https://pytorch.org/ignite/v0.4.8/generated/ignite.handlers.checkpoint.ModelCheckpoint.html)
                class.
        """
        self.kwargs = {
            "filename_prefix": "",
            "global_step_transform": global_step_transform,
            "filename_pattern": "{filename_prefix}{name}_{global_step}.{ext}",
            **kwargs,
        }
        # Create handler here in case needed by load_objects or last_checkpoint
        # before __call__ is used
        self.objs = ModelCheckpoint(**self.kwargs)

        # For saving self.objs. Only save the very latest (n_saved = 1)
        self.ckpter = ModelCheckpoint(**{**self.kwargs, "n_saved": 1})

    def __call__(
        self,
        adapter=None,
        validator=None,
        val_hooks=None,
        **kwargs,
    ):
        """
        Creates the checkpointing function.
        Arguments:
            adapter: An [```Adapter```][pytorch_adapt.adapters.BaseAdapter] object.
            validator: A [```ScoreHistory```][pytorch_adapt.validators.ScoreHistory] object.
            val_hooks: A list of functions called during validation.
                See [```Ignite```][pytorch_adapt.frameworks.ignite.Ignite] for details.
        """
        self.objs = ModelCheckpoint(**{**self.kwargs, **kwargs})
        dict_to_save = {}
        if adapter:
            dict_to_save.update(adapter_to_dict(adapter))
        if validator:
            dict_to_save["validator"] = validator
        if val_hooks:
            dict_to_save.update(val_hooks_to_dict(val_hooks))

        def fn(engine):
            self.objs(engine, {"engine": engine, **dict_to_save})
            self.ckpter(engine, {"checkpointer": self.objs})

        return fn

    def load_objects(self, to_load, checkpoint=None, global_step=None):
        to_load = {k: v for k, v in to_load.items() if v}
        if global_step is not None:
            self.objs.reload_objects(
                to_load, name="checkpoint", global_step=global_step
            )
        else:
            self.objs.load_objects(to_load, str(checkpoint))

    def load_best_checkpoint(self, to_load):
        last_checkpoint = self.get_best_checkpoint()
        self.load_objects(to_load, last_checkpoint)

    def get_best_checkpoint(self):
        if self.objs.last_checkpoint:
            return self.objs.last_checkpoint

        ckpter_last_checkpoint = self.ckpter.last_checkpoint
        if not ckpter_last_checkpoint:
            files = glob.glob(
                os.path.join(self.ckpter.save_handler.dirname, "*checkpointer*.pt")
            )
            if len(files) > 1:
                raise ValueError("there should only be 1 matching checkpointer file")
            ckpter_last_checkpoint = files[0]

        self.ckpter.load_objects(
            {"checkpointer": self.objs}, str(ckpter_last_checkpoint)
        )
        return self.objs.last_checkpoint



import torch
import torchvision.transforms as T

from .constants import IMAGENET_MEAN, IMAGENET_STD


class GrayscaleToRGB:
    def __call__(self, x):
        return torch.cat([x, x, x], dim=0)


def get_mnist_transform(domain, **kwargs):
    if domain == "mnist":
        return T.Compose(
            [
                T.Resize(32),
                T.ToTensor(),
                GrayscaleToRGB(),
                T.Normalize(mean=IMAGENET_MEAN, std=IMAGENET_STD),
            ]
        )
    elif domain == "mnistm":
        return T.Compose(
            [
                T.ToTensor(),
                T.Normalize(mean=IMAGENET_MEAN, std=IMAGENET_STD),
            ]
        )


def get_resnet_transform(is_training, **kwargs):
    transform = [T.Resize(256)]
    if is_training:
        transform += [
            T.RandomCrop(224),
            T.RandomHorizontalFlip(),
        ]
    else:
        transform += [T.CenterCrop(224)]

    transform += [
        T.ToTensor(),
        T.Normalize(mean=IMAGENET_MEAN, std=IMAGENET_STD),
    ]
    return T.Compose(transform)


def get_timm_transform(is_training, **kwargs):
    from timm.data.transforms_factory import create_transform

    return create_transform(
        input_size=224, is_training=is_training, auto_augment="original"
    )



import itertools
from typing import Dict, List, Tuple, Union

from pytorch_metric_learning.utils import common_functions as pml_cf

from ..utils import common_functions as c_f
from .base_validator import BaseValidator


class MultipleValidators(BaseValidator):
    """
    Wraps multiple validators and returns the sum of their scores.
    """

    def __init__(
        self,
        validators: Union[List["BaseValidator"], Dict[str, "BaseValidator"]],
        weights: Union[List[float], Dict[str, float]] = None,
        return_sub_scores=False,
        **kwargs,
    ):
        """
        Arguments:
            validators: A list of validators or a dictionary mapping from strings to validators.
            weights: A list of floats or a dictionary mapping from validator names to floats.
                If ```None```, then the validators will be equally weighted.
            return_sub_scores: If ```True```, then return the score of each validator,
                in addition to their summed value.
        """
        super().__init__(**kwargs)
        self.validators = c_f.enumerate_to_dict(validators)
        self.weights = c_f.default(weights, {k: 1 for k in self.validators.keys()})
        self.weights = c_f.enumerate_to_dict(self.weights)
        if self.validators.keys() != self.weights.keys():
            raise KeyError("validator keys and weight keys must be the same")
        self.return_sub_scores = return_sub_scores
        pml_cf.add_to_recordable_attributes(self, list_of_names=["weights"])

    def _required_data(self):
        output = [v.required_data for v in self.validators.values()]
        output = list(itertools.chain(*output))
        return list(set(output))

    def compute_score(self):
        pass

    def __call__(self, **kwargs) -> Union[float, Tuple[float, Dict[str, float]]]:
        """
        Returns:
            The sum of the validator scores. If ```self.return_sub_scores``` then
                it also returns a dictionary containing each validator's weighted score.
        """
        kwargs = self.kwargs_check(kwargs)
        outputs = {}
        for k, v in self.validators.items():
            score = v(**c_f.filter_kwargs(kwargs, v.required_data))
            outputs[k] = score * self.weights[k]
        final = sum(outputs.values())
        if self.return_sub_scores:
            return final, outputs
        return final

    def __repr__(self):
        return c_f.nice_repr(self, self.extra_repr(), self.validators)

    def extra_repr(self):
        x = super().extra_repr()
        x += f"\n{c_f.extra_repr(self, ['weights'])}"
        return x



from abc import abstractmethod

from ..utils import common_functions as c_f
from .base_validator import BaseValidator


class TorchmetricsValidator(BaseValidator):
    def __init__(self, layer="preds", torchmetric_kwargs=None, **kwargs):
        """
        Arguments:
            layer: The name of the layer to pass into the accuracy function.
                For example, ```"preds"``` refers to the softmaxed logits.
            torchmetric_kwargs: A dictionary of keyword arguments to be
                passed into the
                [torchmetrics accuracy function](https://torchmetrics.readthedocs.io/en/latest/references/functional.html#accuracy-func).
        """
        super().__init__(**kwargs)
        self.layer = layer
        self.torchmetric_kwargs = c_f.default(torchmetric_kwargs, {})

    def compute_score(self, src_val):
        return self.accuracy_fn(
            src_val[self.layer], src_val["labels"], **self.torchmetric_kwargs
        ).item()

    @property
    @abstractmethod
    def accuracy_fn(self):
        pass



import torch
from pytorch_metric_learning.utils import common_functions as pml_cf


class NLLLoss(torch.nn.Module):
    """
    Same as torch.nn.NLLLoss but takes in softmax as input
    """

    def __init__(self, reduction="mean"):
        super().__init__()
        self.reduction = reduction

    def forward(self, x: torch.Tensor, y: torch.Tensor):
        """ """
        x = torch.log(x + pml_cf.small_val(x.dtype))
        return torch.nn.functional.nll_loss(x, y, reduction=self.reduction)



import os

from .base_dataset import BaseDownloadableDataset
from .utils import check_length, check_train


class MNISTM(BaseDownloadableDataset):
    """
    The dataset used in "Domain-Adversarial Training of Neural Networks".
    It consists of colored MNIST digits.

    Extends [BaseDownloadableDataset][pytorch_adapt.datasets.BaseDownloadableDataset],
    so the dataset can be downloaded by setting ```download=True``` when
    initializing.
    """

    url = "https://cornell.box.com/shared/static/jado7quprg6hzzdubvwzh9umr75damwi"
    filename = "mnist_m.tar.gz"
    md5 = "859df31c91afe82e80e5012ba928f279"

    def __init__(self, root: str, train: bool, transform=None, **kwargs):
        """
        Arguments:
            root: The dataset must be located at ```<root>/mnist_m```
            train: Whether or not to use the training set.
            transform: The image transform applied to each sample.
        """
        self.train = check_train(train)
        super().__init__(root=root, domain="MNISTM", **kwargs)
        self.transform = transform

    def set_paths_and_labels(self, root):
        name = "train" if self.train else "test"
        labels_file = os.path.join(root, "mnist_m", f"mnist_m_{name}_labels.txt")
        img_dir = os.path.join(root, "mnist_m", f"mnist_m_{name}")
        with open(labels_file) as f:
            content = [line.rstrip().split(" ") for line in f]
        self.img_paths = [os.path.join(img_dir, x[0]) for x in content]
        check_length(self, {"train": 59001, "test": 9001}[name])
        self.labels = [int(x[1]) for x in content]



import os
from collections import OrderedDict

from .base_dataset import BaseDataset, BaseDownloadableDataset
from .utils import check_img_paths, check_length, check_train


class DomainNet(BaseDataset):
    """
    A large dataset used in "Moment Matching for Multi-Source Domain Adaptation".
    It consists of 345 classes in 6 domains:
    clipart, infograph, painting, quickdraw, real, sketch
    """

    def __init__(self, root: str, domain: str, train: bool, transform, **kwargs):
        """
        Arguments:
            root: The dataset must be located at ```<root>/domainnet```
            domain: One of the 6 domains
            train: Whether or not to use the training set.
            transform: The image transform applied to each sample.
        """
        super().__init__(domain=domain, **kwargs)
        if not isinstance(train, bool):
            raise TypeError("train should be True or False")
        name = "train" if train else "test"
        labels_file = os.path.join(root, "domainnet", f"{domain}_{name}.txt")
        img_dir = os.path.join(root, "domainnet")

        with open(labels_file) as f:
            content = [line.rstrip().split(" ") for line in f]
        self.img_paths = [os.path.join(img_dir, x[0]) for x in content]
        check_img_paths(img_dir, self.img_paths, domain)
        check_length(
            self,
            {
                "clipart": {"train": 33525, "test": 14604}[name],
                "infograph": {"train": 36023, "test": 15582}[name],
                "painting": {"train": 50416, "test": 21850}[name],
                "quickdraw": {"train": 120750, "test": 51750}[name],
                "real": {"train": 120906, "test": 52041}[name],
                "sketch": {"train": 48212, "test": 20916}[name],
            }[domain],
        )
        self.labels = [int(x[1]) for x in content]
        self.transform = transform


class DomainNet126Full(BaseDataset):
    """
    A subset of DomainNet consisting of 126 classes and 4 domains:
    clipart, painting, real, sketch
    """

    def __init__(self, root: str, domain: str, transform, **kwargs):
        """
        Arguments:
            root: The dataset must be located at ```<root>/domainnet```
            domain: One of the 4 domains
            transform: The image transform applied to each sample.
        """
        super().__init__(domain=domain, **kwargs)
        filenames = [
            f"labeled_source_images_{domain}",
            f"labeled_target_images_{domain}_1",
            f"labeled_target_images_{domain}_3",
            f"unlabeled_target_images_{domain}_1",
            f"unlabeled_target_images_{domain}_3",
            f"validation_target_images_{domain}_3",
        ]
        filenames = [os.path.join(root, "domainnet", f"{f}.txt") for f in filenames]
        img_dir = os.path.join(root, "domainnet")

        content = OrderedDict()
        for f in filenames:
            with open(f) as fff:
                for line in fff:
                    path, label = line.rstrip().split(" ")
                    content[path] = label

        self.img_paths = [os.path.join(img_dir, x) for x in content.keys()]
        check_img_paths(img_dir, self.img_paths, domain)
        self.labels = [int(x) for x in content.values()]
        self.transform = transform


class DomainNet126(BaseDownloadableDataset):
    """
    A custom train/test split of DomainNet126Full.
    """

    url = "https://cornell.box.com/shared/static/5uu0v3rs9heusbiht2nn1gbn4yfspas6"
    filename = "domainnet126.tar.gz"
    md5 = "50f29fa0152d715c036c813ad67502d6"

    def __init__(self, root: str, domain: str, train: bool, transform=None, **kwargs):
        """
        Arguments:
            root: The dataset must be located at ```<root>/domainnet```
            domain: One of the 4 domains
            train: Whether or not to use the training set.
            transform: The image transform applied to each sample.
        """
        self.train = check_train(train)
        super().__init__(root=root, domain=domain, **kwargs)
        self.transform = transform

    def set_paths_and_labels(self, root):
        name = "train" if self.train else "test"
        labels_file = os.path.join(root, "domainnet", f"{self.domain}126_{name}.txt")
        img_dir = os.path.join(root, "domainnet")

        with open(labels_file) as f:
            content = [line.rstrip().split(" ") for line in f]
        self.img_paths = [os.path.join(img_dir, x[0]) for x in content]
        check_img_paths(img_dir, self.img_paths, self.domain)
        check_length(
            self,
            {
                "clipart": {"train": 14962, "test": 3741}[name],
                "painting": {"train": 25201, "test": 6301}[name],
                "real": {"train": 56286, "test": 14072}[name],
                "sketch": {"train": 19665, "test": 4917}[name],
            }[self.domain],
        )
        self.labels = [int(x[1]) for x in content]



from ..hooks import CDANNEHook, DANNEHook, DANNHook, GVBEHook, GVBHook
from ..layers import ModelWithBridge
from .base_adapter import BaseGCDAdapter
from .gan import CDAN
from .utils import with_opt


class DANN(BaseGCDAdapter):
    """
    Wraps [DANNHook][pytorch_adapt.hooks.DANNHook].

    |Container|Required keys|
    |---|---|
    |models|```["G", "C", "D"]```|
    |optimizers|```["G", "C", "D"]```|
    """

    def init_hook(self, hook_kwargs):
        opts = with_opt(list(self.optimizers.keys()))
        self.hook = self.hook_cls(opts=opts, **hook_kwargs)

    @property
    def hook_cls(self):
        return DANNHook


class DANNE(DANN):
    @property
    def hook_cls(self):
        return DANNEHook


class CDANNE(DANN, CDAN):
    @property
    def hook_cls(self):
        return CDANNEHook


class GVB(DANN):
    """
    Wraps [GVBHook][pytorch_adapt.hooks.GVBHook].

    |Container|Required keys|
    |---|---|
    |models|```["G", "C", "D"]```|
    |optimizers|```["G", "C", "D"]```|

    Models D and C must be of type [```ModelWithBridge```][pytorch_adapt.layers.ModelWithBridge].
    If not, they will be converted into instances of ```ModelWithBridge```,
    with each bridge being a re-initialized copy of each model.
    """

    @property
    def hook_cls(self):
        return GVBHook

    def init_containers_and_check_keys(self, containers):
        models = containers["models"]
        for k in ["D", "C"]:
            if not isinstance(models[k], ModelWithBridge):
                models[k] = ModelWithBridge(models[k])
        super().init_containers_and_check_keys(containers)


class GVBE(GVB):
    @property
    def hook_cls(self):
        return GVBEHook
